#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

from q2k.reference import *
from q2k.core import *
from q2k.gui import *




class ref:
    # QMK -> Keyplus Keycode Mapping
    # for keycodes only, functions are in KEYP_FUNC_LIST
    keyp_kc = {
        "KC_NO"                            :          " no ", #- giorgio, 2018
        "KC_TRNS"                          :          "____",
        "KC_TRANSPARENT"                   :          "____", #"trans",
        "XXXXXXX"                          :          " no ",
        "_______"                          :          "____",
        "RESET"                            :          "boot",
        "MAGIC_HOST_NKRO"                  :          "kro_n",                       # FORCE NKRO
        "MAGIC_UNHOST_NKRO"                :          "kro_6",                       # FORCE 6KRO
        "MAGIC_TOGGLE_NKRO"                :          "kro_auto",                    # Set Auto NKRO
        "KC_GESC"                          :          "gesc",

        # Alphas
        "KC_GRV"                           :          "'`' ",
        "KC_GRAVE"                         :          "'`' ",  #"grave",             # Grave (~`)kro_6
        "KC_1"                             :          " 1 ",
        "KC_2"                             :          " 2 ",
        "KC_3"                             :          " 3 ",
        "KC_4"                             :          " 4 ",
        "KC_5"                             :          " 5 ",
        "KC_6"                             :          " 6 ",
        "KC_7"                             :          " 7 ",
        "KC_8"                             :          " 8 ",
        "KC_9"                             :          " 9 ",
        "KC_0"                             :          " 0 ",
        "KC_MINS"                          :          "'-' ",
        "KC_MINUS"                         :          "'-' ", #"minus",              # Minus (-_)
        "KC_EQL"                           :          "'=' ",
        "KC_EQUAL"                         :          "'=' ", #"equal",              # Equal (=+)
        "KC_Q"                             :          " q  ",
        "KC_W"                             :          " w  ",
        "KC_E"                             :          " e  ",
        "KC_R"                             :          " r  ",
        "KC_T"                             :          " t  ",
        "KC_Y"                             :          " y  ",
        "KC_U"                             :          " u  ",
        "KC_I"                             :          " i  ",
        "KC_O"                             :          " o  ",
        "KC_P"                             :          " p  ",
        "KC_LBRC"                          :          "'[' ",
        "KC_LBRACKET"                      :          "'[' ", #"left_bracket",        # Left Bracket ([)
        "KC_RBRC"                          :          "']' ",
        "KC_RBRACKET"                      :          "']' ", #"right_bracket",       # Right Bracket (])
        "KC_BSLS"                          :          "'\\' ", #"bsls",
        "KC_BSLASH"                        :          "'\\' ",                        # Backslash (\|)
        "KC_A"                             :          " a  ",
        "KC_S"                             :          " s  ",
        "KC_D"                             :          " d  ",
        "KC_F"                             :          " f  ",
        "KC_G"                             :          " g  ",
        "KC_H"                             :          " h  ",
        "KC_J"                             :          " j  ",
        "KC_K"                             :          " k  ",
        "KC_L"                             :          " l  ",
        "KC_SCLN"                          :          "';' ",
        "KC_SCOLON"                        :          "';' ", #"semicolon",          # Semicolon (;:)
        "KC_QUOT"                          :          "\"\'\" ", 
        "KC_QUOTE"                         :          "\"\'\" ", #"quot",            # Quote ('")
        "KC_Z"                             :          " z  ",
        "KC_X"                             :          " x  ",
        "KC_C"                             :          " c  ",
        "KC_V"                             :          " v  ",
        "KC_B"                             :          " b  ",
        "KC_N"                             :          " n  ",
        "KC_M"                             :          " m  ",
        "KC_COMM"                          :          "',' ",
        "KC_COMMA"                         :          "',' ", #"comma",              # Comma (,<)
        "KC_DOT"                           :          "'.' ",                        # Period (.>)
        "KC_SLSH"                          :          "'/' ",
        "KC_SLASH"                         :          "'/' ", #"forward_slash",      # Forward Slash (/?)
        "KC_SPC"                           :          " spc",
        "KC_SPACE"                         :          " spc", #"spacebar",           # Spacebar

        # Modifiers
        "KC_RCTL"                          :          "rctl",
        "KC_RCTRL"                         :          "rctl", #"rctrl",              # Right Ctrl
        "KC_LCTL"                          :          "lctl",
        "KC_LCTRL"                         :          "lctl", #"lctrl",              # Left Ctrl
        "KC_RSFT"                          :          "rsft",
        "KC_RSHIFT"                        :          "rsft", #"rshift",             # Right Shift
        "KC_LSFT"                          :          "lsft",
        "KC_LSHIFT"                        :          "lsft", #"lshift",             # Left Shift
        "KC_LGUI"                          :          "lgui",
        "KC_LWIN"                          :          "lgui", #"lwin",
        "KC_LCMD"                          :          "lgui",                        # Left GUI/Win
        "KC_RGUI"                          :          "rgui",
        "KC_RWIN"                          :          "rgui", #"rwin",                    
        "KC_RCMD"                          :          "rgui",                        # Right GUI/Win
        "KC_LALT"                          :          "lalt",                        # Left Alt
        "KC_RALT"                          :          "ralt",                        # Right Alt
        "KC_BSPC"                          :          "bspc",
        "KC_BSPACE"                        :          "bspc", #"backspace",          # Backspace
        "KC_ENT"                           :          " ent",
        "KC_ENTER"                         :          " ent", #"enter",              # Enter
        "KC_TAB"                           :          " tab",                        # Tab
        "KC_CAPS"                          :          "caps",
        "KC_CLCK"                          :          "caps",
        "KC_CAPSLOCK"                      :          "caps", #"caps_lock""kp_ent",, # Caps Lock
        "KC_RGHT"                          :          "rght",
        "KC_RIGHT"                         :          "rght", #"right",              # right arrow
        "KC_UP"                            :          " up ",
        "KC_DOWN"                          :          "down",
        "KC_LEFT"                          :          "left",
        # Function 
        "KC_ESC"                           :          " esc",
        "KC_ESCAPE"                        :          " esc", #"escape",             # Escape
        "KC_F1"                            :          " f1 ",

        "KC_F2"                            :          " f2 ",
        "KC_F3"                            :          " f3 ",
        "KC_F4"                            :          " f4 ",
        "KC_F5"                            :          " f5 ",
        "KC_F6"                            :          " f6 ",
        "KC_F7"                            :          " f7 ",
        "KC_F8"                            :          " f8 ",
        "KC_F9"                            :          " f9 ",
        "KC_F10"                           :          " f10",
        "KC_F11"                           :          " f11",
        "KC_F12"                           :          " f12",
        "KC_PSCR"                          :          "pscr",
        "KC_PSCREEN"                       :          "pscr", #"print_screen",       # Print Screen
        "KC_SLCK"                          :          "slck",
        "KC_SCROLLLOCK"                    :          "slck", #"scroll_lock",        # Scroll Lock
        "KC_PAUS"                          :          "paus", 
        "KC_BRK"                           :          "paus",
        "KC_PAUSE"                         :          "paus", #"pause",              # Pause/Break
        "KC_INS"                           :          " ins",
        "KC_INSERT"                        :          " ins", #"insert",             # Insert
        "KC_DEL"                           :          " del",
        "KC_DELT"                          :          " del",
        "KC_DELETE"                        :          " del", #"delete",             # Delete
        "KC_HOME"                          :          "home",                        # Home
        "KC_END"                           :          " end",                        # End
        "KC_PGUP"                          :          "pgup",                        # Page Up
        "KC_PGDN"                          :          "pgdn",
        "KC_PGDOWN"                        :          "pgdn", #"page_down",          # Page Down
        "KC_F13"                           :          " f13",                     
        "KC_F14"                           :          " f14",
        "KC_F15"                           :          " f15",
        "KC_F16"                           :          " f16",
        "KC_F17"                           :          " f17",
        "KC_F18"                           :          " f18",
        "KC_F19"                           :          " f19",
        "KC_F20"                           :          " f20",
        "KC_F21"                           :          " f21",
        "KC_F22"                           :          " f22",
        "KC_F23"                           :          " f23",
        "KC_F24"                           :          " f24",
        # Shifted
        "KC_TILDE"                         :          "'~' ",
        "KC_TILD"                          :          "'~' ",
        "KC_EXCLAIM"                       :          "'!' ",
        "KC_EXLM"                          :          "'!' ",
        "KC_AT"                            :          "'@' ",
        "KC_HASH"                          :          "'#' ",
        "KC_DOLLAR"                        :          "'$' ",
        "KC_DLR"                           :          "'$' ",
        "KC_PERCENT"                       :          "'%' ",
        "KC_PERC"                          :          "'%' ",
        "KC_CIRCUMFLEX"                    :          "'^' ",
        "KC_CIRC"                          :          "'^' ",
        "KC_AMPERSAND"                     :          "'&' ",
        "KC_AMPR"                          :          "'&' ",
        "KC_ASTERISK"                      :          "'*' ",
        "KC_ASTR"                          :          "'*' ",
        "KC_LEFT_PAREN"                    :          "'(' ",
        "KC_LPRN"                          :          "'(' ",
        "KC_RIGHT_PAREN"                   :          "')' ",
        "KC_RPRN"                          :          "')' ",
        "KC_UNDERSCORE"                    :          "'_' ",
        "KC_UNDS"                          :          "'_' ",
        "KC_PLUS"                          :          "'+' ",
        "KC_LEFT_CURLY_BRACE"              :          "'{' ",
        "KC_LCBR"                          :          "'{' ",
        "KC_RIGHT_CURLY_BRACE"             :          "'}' ",
        "KC_RCBR"                          :          "'}' ",
        "KC_PIPE"                          :          "'|' ",
        "KC_COLON"                         :          "':' ",
        "KC_COLN"                          :          "':' ",
        "KC_DOUBLE_QUOTE"                  :          "\'\"\' ",
        "KC_DQT"                           :          "\'\"\' ",
        "KC_DQUO"                          :          "\'\"\' ",
        "KC_LEFT_ANGLE_BRACKET"            :          "'<' ",
        "KC_LT"                            :          "'<' ",
        "KC_LABK"                          :          "'<' ",
        "KC_RIGHT_ANGLE_BRACKET"           :          "'>' ",
        "KC_GT"                            :          "'>' ",
        "KC_RABK"                          :          "'>' ",
        "KC_QUESTION"                      :          "'?' ",
        "KC_QUES"                          :          "'?' ",

        # Numpad
        "KC_NLCK"                          :          "nlck", 
        "KC_NUMLOCK"                       :          "nlck", #"num_lock",           # Num Lock
        "KC_P1"                            :          "kp_1",
        "KC_KP_1"                          :          "kp_1",                        # Numpad 1
        "KC_P2"                            :          "kp_2",   
        "KC_KP_2"                          :          "kp_2",                        # Numpad 2
        "KC_P3"                            :          "kp_3",   
        "KC_KP_3"                          :          "kp_3",                        # Numpad 3
        "KC_P4"                            :          "kp_4",   
        "KC_KP_4"                          :          "kp_4",                        # Numpad 4
        "KC_P5"                            :          "kp_5",   
        "KC_KP_5"                          :          "kp_5",                        # Numpad 5
        "KC_P6"                            :          "kp_6",   
        "KC_KP_6"                          :          "kp_6",                        # Numpad 6
        "KC_P7"                            :          "kp_7",   
        "KC_KP_7"                          :          "kp_7",                        # Numpad 7
        "KC_P8"                            :          "kp_8",   
        "KC_KP_8"                          :          "kp_8",                        # Numpad 8
        "KC_P9"                            :          "kp_9",   
        "KC_KP_9"                          :          "kp_9",                        # Numpad 9
        "KC_P0"                            :          "kp_0",   
        "KC_KP_0"                          :          "kp_0",                        # Numpad 0
        "KC_PDOT"                          :          "kp_.",
        "KC_KP_DOT"                        :          "kp_.",                        # Numpad .
        "KC_PCMM"                          :          "kp_,",
        "KC_KP_COMMA"                      :          "kp_,",                        # Numpad ,
        "KC_PSLS"                          :          "kp_/",
        "KC_KP_SLASH"                      :          "kp_/",                        # Numpad /
        "KC_PAST"                          :          "kp_*",                      
        "KC_KP_ASTERISK"                   :          "kp_*",                        # Numpad *
        "KC_PMNS"                          :          "kp_-", 
        "KC_KP_MINUS"                      :          "kp_-",                        # Numpad -
        "KC_PPLS"                          :          "kp_+",
        "KC_KP_PLUS"                       :          "kp_+",                        # Numpad +
        "KC_PEQL"                          :          "kp_=",
        "KC_KP_EQUAL"                      :          "kp_=",                        # Numpad =
        "KC_PENT"                          :          "kp_ent",
        "KC_KP_ENTER"                      :          "kp_ent", #"kp_enter",         # Numpad Enter

        # Misc Functions
        "KC_APP"                           :          " app",
        "KC_APPLICATION"                   :          " app",                        # Application
        "KC_LCAP"                          :          "locking_caps_lock",
        "KC_EXEC"                          :          "execute",
        "KC_EXECUTE"                       :          "execute",                     # Execute
        "KC_SLCT"                          :          "select",
        "KC_SELECT"                        :          "select",                      # Select____
        "KC_AGIN"                          :          "again", 
        "KC_AGAIN"                         :          "again",                       # Again
        "KC_HELP"                          :          "help",
        "KC_MENU"                          :          "hid_menu",                    # Menu
        "KC_UNDO"                          :          "undo",                        # Undo
        "KC_CUT"                           :          " cut",                        # Cut
        "KC_COPY"                          :          "copy",                        # Copy
        "KC_PSTE"                          :          "paste", 
        "KC_PASTE"                         :          "paste",                       # Paste
        "KC_FIND"                          :          "find",                        # Find
        "KC_ALT_ERASE"                     :          "alternate_erase",             # Alt Erase
        "KC_CANCEL"                        :          "cancel",                      # Cancel
        "KC_SYSREQ"                        :          "sys_req",                     # SYSREQ
        "KC_PRIOR"                         :          "prior",                   
        "KC_SEPERATOR"                     :          "separator",               
        "KC_RETURN"                        :          "return",                   
        "KC_OUT"                           :          " out",                      
        "KC_OPER"                          :          "oper",                     
        "KC_CLEAR_AGAIN"                   :          "clear_again",              
        "KC_CRSEL"                         :          "crsel",
        "KC_EXSEL"                         :          "exsel",
        "KC_STOP"                          :          "stop",

        "KC_LOCKING_CAPS"                  :          "locking_caps_lock",           # Locking Caps Lock
        "KC_LNUM"                          :          "locking_num_lock",
        "KC_LOCKING_NUM"                   :          "locking_num_lock",            # Locking Num Lock
        "KC_LSCR"                          :          "locking_scroll_lock",
        "KC_LOCKING_SCROLL"                :          "locking_scroll_lock",         # Locking Scroll Lock
        "KC_ERAS"                          :          "alternate_erase",
        "KC_ALT_ERASE"                     :          "alternate_erase",             # Alternate Erase
        "KC_CLR"                           :          "clear", 
        "KC_CLEAR"                         :          "clear",                       # Clear
        "KC_NUHS"                          :          "iso#",
        "KC_NONUS_HASH"                    :          "iso_hash",                    # ISO hash (#~)
        "KC_NUBS"                          :          "iso\\",
        "KC_NONUS_BSLASH"                  :          "iso\\",                       # ISO Backslash (\|)
        "KC_ZKHK"                          :          "'`' ",                        # JIS Grave
        "KC_RO"                            :          "int1",          
        "KC_INT1"                          :          "int1",                        # JIS \|
        "KC_KANA"                          :          "int2",
        "KC_INT2"                          :          "int2",                        # JIS Katakana/Hiragana
        "KC_JYEN"                          :          "int3",
        "KC_INT3"                          :          "int3",                        # JIS Y
        "KC_HENK"                          :          "int4",
        "KC_INT4"                          :          "int4",                        # JIS Henkan
        "KC_MHEN"                          :          "int5",
        "KC_INT5"                          :          "int5",                        # JIS Muhenkan
        "KC_INT6"                          :          "int6",
        "KC_INT7"                          :          "int7",
        "KC_INT8"                          :          "int8",
        "KC_INT9"                          :          "int9",
        "KC_HAEN"                          :          "lang1",
        "KC_LANG1"                         :          "lang1",                      # KR Hangul/ENG
        "KC_HANJ"                          :          "lang2",
        "KC_LANG2"                         :          "lang2",                      # KR Hanja
        "KC_LANG3"                         :          "lang3",
        "KC_LANG4"                         :          "lang4",
        "KC_LANG5"                         :          "lang5",
        "KC_LANG6"                         :          "lang6",
        "KC_LANG7"                         :          "lang7",
        "KC_LANG8"                         :          "lang8",
        "KC_LANG9"                         :          "lang9",
     

        # Media Controls
        "KC__MUTE"                         :          "mute",
        "KC_MUTE"                          :          "mute",
        "KC_AUDIO_MUTE"                    :          "mute", #"audio_mute",         # Audio Mute
        "KC_VOLU"                          :          "volu",
        "KC__VOLUP"                        :          "volu",
        "KC_AUDIO_VOL_UP"                  :          "volu", #"audio_vol_up",       # Audio Vol. Up
        "KC_VOLD"                          :          "vold", 
        "KC__VOLDOWN"                      :          "vold",
        "KC_AUDIO_VOL_DOWN"                :          "vold", #"audio_vol_down",     # Audio Vol. Down
        "KC_MNXT"                          :          "mnxt",
        "KC_MEDIA_NEXT_TRACK"              :          "mnxt", #"media_next_track",   # Media Next Track
        "KC_MPRV"                          :          "mprv",
        "KC_MEDIA_PREV_TRACK"              :          "mprv", #"media_prev_track",   # Media Prev Track
        "KC_MFFD"                          :          "mffd",
        "KC_MEDIA_FAST_FORWARD"            :          "mffd",#"media_fast_forward",  # Media Fast Forward
        "KC_MRWD"                          :          "mrwd",
        "KC_MEDIA_REWIND"                  :          "mrwd", #"media_rewind",       # Media Rewind
        "KC_MSTP"                          :          "mstp",
        "KC_MEDIA_STOP"                    :          "mstp", #"media_stop",         # Media Stop
        "KC_MPLY"                          :          "mply",
        "KC_MEDIA_PLAY_PAUSE"              :          "mply", #"media_play_pause",   # Media Play/Pause
        "KC_MSEL"                          :          "msel",
        "KC_MEDIA_SELECT"                  :          "msel", #"media_select",       # Media Select
        "KC_EJCT"                          :          "mjct",
        "KC_MEDIA_EJECT"                   :          "mjct", #"media_eject",        # Media Eject

        # WWW
        "KC_MAIL"                          :          "mail",                        # Mail
        "KC_CALC"                          :          "calc",
        "KC_CALCULATOR"                    :          "calc",                        # Calculator
        "KC_MYCM"                          :          "comp",
        "KC_MY_COMPUTER"                   :          "my_computer",                 # My Computer
        "KC_WSCH"                          :          "www_search",
        "KC_WWW_SEARCH"                    :          "www_search",                  # WWW Search
        "KC_WHOM"                          :          "www_home", 
        "KC_WWW_HOME"                      :          "www_home",                    # WWW Home
        "KC_WBAK"                          :          "www_back", 
        "KC_WWW_BACK"                      :          "www_back",                    # WWW Back
        "KC_WFWD"                          :          "www_forward",
        "KC_WWW_FORWARD"                   :          "www_forward",                 # WWW Forward
        "KC_WSTP"                          :          "www_stop",
        "KC_WWW_STOP"                      :          "www_stop",                    # WWW Stop
        "KC_WREF"                          :          "www_refresh",
        "KC_WWW_REFRESH"                   :          "www_refresh",                 # WWW Refresh
        "KC_WFAV"                          :          "www_favourites",
        "KC_WWW_FAVORITES"                 :          "www_favourites",              # WWW Favourites
        
        # Mousekey
        "KC_MS_U"                          :          "ms_u",
        "KC_MS_UP"                         :          "ms_u", #"mouse_up",           # Mouse Up
        "KC_MS_D"                          :          "ms_d",
        "KC_MS_DOWN"                       :          "ms_d", #"mouse_down",         # Mouse Down
        "KC_MS_L"                          :          "ms_l", 
        "KC_MS_LEFT"                       :          "ms_l", #"mouse_left",         # Mouse Left
        "KC_MS_R"                          :          "ms_r", 
        "KC_MS_RIGHT"                      :          "ms_r", #"mouse_right",        # Mouse Right
        "KC_BTN1"                          :          "btn1",
        "KC_MS_BTN1"                       :          "btn1", #"mouse_btn1",         # MOUSE1
        "KC_BTN2"                          :          "btn2", 
        "KC_MS_BTN2"                       :          "btn2", #"mouse_btn2",         # MOUSE2
        "KC_BTN3"                          :          "btn3",
        "KC_MS_BTN3"                       :          "btn3", #"mouse_btn3",         # MOUSE3
        "KC_BTN4"                          :          "btn4",
        "KC_MS_BTN4"                       :          "btn4", #"mouse_btn4",         # MOUSE4
        "KC_BTN5"                          :          "btn5",
        "KC_MS_BTN5"                       :          "btn5", #"mouse_btn5",         # MOUSE5
        "KC_WH_U"                          :          "wh_u",
        "KC_MS_WH_UP"                      :          "wh_u", #"mouse_wh_up",        # Mouse Wheel Up
        "KC_WH_D"                          :          "wh_d",
        "KC_MS_WH_DOWN"                    :          "wh_d", #"mouse_wh_down",      # Mouse Wheel Down
        "KC_WH_L"                          :          "wh_l", 
        "KC_MS_WH_LEFT"                    :          "wh_l", #"mouse_wh_left",      # Mouse Wheel Left
        "KC_WH_R"                          :          "wh_r", 
        "KC_MS_WH_RIGHT"                   :          "wh_r", #"mouse_wh_right",     # Mouse Wheel Right

        # System
        "KC_PWR"                           :          "system_power",
        "KC_POWER"                         :          "system_power",
        "KC_SYSTEM_POWER"                  :          "system_power",                # Power
        "KC_SLEP"                          :          "system_sleep",
        "KC_SYSTEM_SLEEP"                  :          "system_sleep",                # Sleep
        "KC_WAKE"                          :          "system_wake",
        "KC_SYSTEM_WAKE"                   :          "system_wake",                 # Wake

        # QMK Modifiers
        "KC_HYPR"                          :          "csag-none",                   # Hyper
        "KC_MEH"                           :          "csa-none",                    # Ctrl+Shift+Alt
    }                           
                              
    # QMK -> Keyplus Layer Functions
    keyp_layer_func = {

        "MO("                              :          " L",
        "TG("                              :          "tog_L",
        "TT("                              :          " L",                          # Feature not fully implemented in K+
        "OSL("                             :          "s_L",
        "TO("                              :          "set_L",
        "DF("                              :          "set_L",                       # Feature not fully implemented in K+

    }                      
    # QMK -> Keyplus functions
    keyp_func = {
        # Mod Keys
        "LCTL("                            :          "C",
        "LSFT("                            :          "S",
        "S("                               :          "S",
        "LALT("                            :          "A",
        "LGUI("                            :          "G",
        "LCMD("                            :          "G",
        "LWIN("                            :          "G",
        "RCTL("                            :          "rC",
        "RSFT("                            :          "rS",
        "RALT("                            :          "rA",
        "RGUI("                            :          "rG",
        "RWIN("                            :          "rG",
        "RCMD("                            :          "rG",
        # Multi Mod Keys
        "SGUI("                            :          "SG",
        "SCMD("                            :          "SG",
        "SWIN("                            :          "SG",
        "LCA("                             :          "CA",
        "ALTG("                            :          "rCrA",
        "LCAG("                            :          "CAG",
        "MEH("                             :          "CSA",
        "HYPR("                            :          "CSAG",

        "OSM(MOD_LSFT)"                    :          "s_lsft",
        "OSM(MOD_LCTL)"                    :          "s_lctrl",
        "OSM(MOD_LALT)"                    :          "s_lalt",
        "OSM(MOD_LGUI)"                    :          "s_lgui",
        "OSM(MOD_RCTL)"                    :          "s_rctrl",
        "OSM(MOD_RSFT)"                    :          "s_rsft",
        "OSM(MOD_RALT)"                    :          "s_ralt",
        "OSM(MOD_RGUI)"                    :          "s_rgui",
        "OSM(KC_LSFT)"                     :          "s_lsft",
        "OSM(KC_LSHIFT)"                   :          "s_lsft",
        "OSM(KC_LCTL)"                     :          "s_lctrl",
        "OSM(KC_LCTRL)"                    :          "s_lctrl",
        "OSM(KC_LALT)"                     :          "s_lalt",
        "OSM(KC_LGUI)"                     :          "s_lgui",
        "OSM(KC_LWIN)"                     :          "s_lgui",
        "OSM(KC_LCMD)"                     :          "s_lgui",
        "OSM(KC_RSFT)"                     :          "s_rsft",
        "OSM(KC_RSHIFT)"                   :          "s_rsft",
        "OSM(KC_RCTL)"                     :          "s_rctrl",
        "OSM(KC_RCTRL)"                    :          "s_rctrl",
        "OSM(KC_RALT)"                     :          "s_ralt",
        "OSM(KC_RGUI)"                     :          "s_rgui",
        "OSM(KC_RWIN)"                     :          "s_rgui",
        "OSM(KC_RCMD)"                     :          "s_rgui", 
    }

    keyp_tap_layer = {
        "LT("                              :          "L",
        "LM("                              :          "L",
    }

    keyp_tap_mod = {
        "LCTL_T("                          :          "C",
        "CTL_T("                           :          "C",
        "RCTL_T("                          :          "rC",
        "LSFT_T("                          :          "S",
        "SFT_T("                           :          "S",
        "RSFT_T("                          :          "rS",
        "LALT_T("                          :          "A",
        "ALT_T("                           :          "A",
        "RALT_T("                          :          "rA",
        "ALGR_T"                           :          "rA",
        "LGUI_T("                          :          "G",
        "LCMD_T("                          :          "G",
        "GUI_T("                           :          "G",
        "RGUI_T("                          :          "rG",
        "RCMD_T("                          :          "rG",
        "RWIN_T("                          :          "rG",
        "LCA_T("                           :          "CA",
        "C_S_T("                           :          "CS",
        "MEH_T("                           :          "CSA",
        "LCAG_T("                          :          "CAG",
        "RCAG_T("                          :          "rCrArG",
        "ALL_T("                           :          "CSAG",
        "SCMD_T("                          :          "SG",
    }

    # Legacy QMK functions
    qmk_legacy_func = ["KC_FN", "FUNC(", "F("]
    keyp_actions = { 
        "ACTION_LAYER_MOMENTARY("          :          " L", 
        "ACTION_LAYER_TOGGLE("             :          "tog_L",
        "ACTION_LAYER_SET("                :          "set_L",
        "ACTION_LAYER_TAP_TOGGLE("         :          "s_L",
        "ACTION_FUNCTION(SHIFT_ESC)"       :          "gesc",
        "ACTION_FUNCTION(ESC_GRV)"         :          "gesc",
        "ACTION_FUNCTION(LFK_ESC_TILDE)"   :          "gesc",
        "ACTION_FUNCTION(ESCAPE)"          :          "gesc",
       #"ACTION_LAYER_TAP_KEY(2, KC_CAPS)"
       #"ACTION_MODS_KEY"
       #"ACTION"
    }

    qmk_legacy_mod = {
        "MOD_LSFT"                         :          "S",
        "MOD_LCTL"                         :          "C",
        "MOD_LALT"                         :          "A",
        "MOD_LGUI"                         :          "G",
        "MOD_RCTL"                         :          "rC",
        "MOD_RSFT"                         :          "rS",
        "MOD_RALT"                         :          "rA",
        "MOD_RGUI"                         :          "rG",
    }

    # QMK Keycodes currently not supported
    keyp_missing = {
        # Setting Mouse Acceleration
        "KC_ACL0", "KC_MS_ACCEL0", "KC_ACL1", "KC_MS_ACCEL1", "KC_ACL2", "KC_MS_ACCEL2",
        # Thermal Printer
        "PRINT_ON", "PRINT_OFF",
        # Arr Gee Bee
        "RGB_TOG", "RGB_VAI", "RGB_VAD", "RGB_SAI", "RGB_MOD", "RGB_HUD", "RGB_SAD", "RGB_HUI",
        # Backlight
        "BL_STEP",
        # QMK Modifiers
        "LCTL(", "LALT(", "LGUI(", "LCMD(", "LWIN(", "RCTL(", "RSFT(", "RALT(", 
        "RGUI(", "RCMD(", "RWIN(", "KC_HYPR", "KC_MEH", "HYPR(", "MEH(", "LCAG(", 
        "ALTG(", "SGUI(", "SCMD(", "SWIN(", "LCA(", "LCTL_T(", "CTL_T(", "RCTL_T(",
        "LSFT_T(", "RSFT_T(", "LALT_T(", "ALT_T(", "RALT_T(", "ALGR_T(", "LGUI_T(",
        "LCMD_T(", "RWIN_T(", "GUI_T(", "RGUI_T(", "RCMD_T(", "RWIN_T(", "C_S_T(",
        "MEH_T(", "LCAG_T(" , "RCAG_T(", "ALL_T(", "SCMD_T(", "SWIN_T(", "LCA_T(",
    }

    # KBfirmware json compatability list
    kbf_kc = [
        "KC_1", "KC_2", "KC_3", "KC_4", "KC_5", "KC_6", "KC_7",
        "KC_8", "KC_9", "KC_0",
        "KC_A", "KC_B", "KC_C", "KC_D", "KC_E", "KC_F", "KC_G",
        "KC_H", "KC_I", "KC_J", "KC_K", "KC_L", "KC_M", "KC_N",
        "KC_O", "KC_P", "KC_Q", "KC_R", "KC_S", "KC_T", "KC_U",
        "KC_V", "KC_W", "KC_X", "KC_Y", "KC_Z",
        "KC_NUHS", "KC_NUBS", "KC_MINS", "KC_EQL", "KC_LBRC", 
        "KC_RBRC", "KC_BSLS", "KC_SCLN", "KC_QUOT", "KC_GRV", 
        "KC_COMM", "KC_DOT",  "KC_SLSH", 
        "KC_ENT", "KC_ESC", "KC_BSPC", "KC_TAB", "KC_SPC", 
        "KC_CAPS", "KC_APP",
        "KC_LCTL", "KC_LSFT", "KC_LALT", "KC_LGUI", "KC_RCTL",
        "KC_RSFT", "KC_RALT", "KC_RGUI", "KC_TRNS", "KC_NO",
        "RESET",
        "KC_EXLM", "KC_AT", "KC_HASH", "KC_DLR", "KC_PERC",
        "KC_CIRC", "KC_AMPR", "KC_ASTR", "KC_LPRN", "KC_RPRN",
        "KC_UNDS", "KC_PLUS", "KC_LCBR", "KC_RCBR", "KC_PIPE",
        "KC_COLN", "KC_DQUO", "KC_TILD", "KC_LABK", "KC_RABK",
        "KC_QUES",
        "KC_F1", "KC_F2", "KC_F3", "KC_F4", "KC_F5", "KC_F6",
        "KC_F7", "KC_F8", "KC_F9", "KC_F10", "KC_F11", "KC_F12",
        "KC_F13", "KC_F14", "KC_F15", "KC_F16", "KC_F17",
        "KC_F18", "KC_F19", "KC_F20", "KC_F21", "KC_F22",
        "KC_F23", "KC_F24",
        "KC_PSCR", "KC_SLCK", "KC_PAUS",
        "KC_INS", "KC_DEL", "KC_HOME", "KC_END", "KC_PGUP", 
        "KC_PGDN", "KC_LEFT", "KC_DOWN", "KC_UP", "KC_RGHT",
        "KC_PWR", "KC_SLEP", "KC_WAKE", "KC_MUTE", "KC_VOLU",
        "KC_VOLD", "KC_MPLY", "KC_MSTP", "KC_MPRV", "KC_MNXT",
        "KC_RO", "KC_KANA", "KC_JYEN", "KC_HENK", "KC_MHEN",
        "KC_LOCKING_CAPS", "KC_LOCKING_NUM", "KC_LOCKING_SCROLL",
        "KC_APP", "KC_CALC", "KC_MYCM", "KC_WSCH", "KC_WHOM",
        "KC_WBAK", "KC_WFWD", "KC_WSTP", "KC_WREF", "KC_WFAV"
        "KC_MSEL",
        "KC_NLCK", "KC_PSLS", "KC_PAST", "KC_PMNS", "KC_PPLS",
        "KC_PDOT", "KC_PEQL", "KC_PENT", "KC_PCMM",
        "KC_P1", "KC_P2", "KC_P3", "KC_P4", "KC_P5", "KC_P6",
        "KC_P7", "KC_P8", "KC_P9", "KC_P0",
        "BL_TOGG", "BL_DEC", "BL_INC", "BL_STEP",
        "RGB_TOG", "RGB_MOD", "RGB_HUI", "RGB_HUD", "RGB_SAI",
        "RGB_SAD",  "RGB_VAI", "RGB_VAD",
        "KC_MS_U", "KC_MS_D", "KC_MS_L", "KC_MS_R", 
        "KC_BTN1", "KC_BTN2", "KC_BTN3", "KC_BTN4", "KC_BTN5",
        "KC_WH_U", "KC_WH_D", "KC_WH_L", "KC_WH_R", 
        "KC_ACL0", "KC_ACL1", "KC_ACL2",
    ]

    kbf_func = [

        "LCTL(", "LSFT(", "LALT(", "LGUI(", "RTL(", "RSFT(",
        "RALT(", "RGUI(", "HYPR(", "MEH(", "LCAG(", "ALTG(", 
        "LT(", "TO(", "MO(", "DF(", "TG(", "OSL(", "OSM(", 
        "MT(", "CTL_T(", "SFT_T(", "ALT_T(", "GUI_T(", 
        "C_S_T(", "MEH_T(", "LCAG_T(", "ALL_T(", "M("
    ]

    kbf_transcode = {

        "S(":"LSFT(", "RCMD(":"RGUI(", "RWIN(":"RGUI(", "LCTL_T(":"CTL_T(", "RCTL_T(":"CTL_T(", "LSFT_T(":"SFT_T(",
        "RSFT_T(":"SFT_T(", "LALT_T(":"ALT_T(", "RALT_T(":"ALT_T(", "LGUI_T(":"GUI_T(", "LCMD_T(":"GUI_T(",
        "LWIN_T(":"GUI_T(", "RGUI_T(":"GUI_T(", "RCMD_T(":"GUI_T(", "RWIN_T(":"GUI_T(", "RCAG_T(":"LCAG_T(",

        "TT(":"MO(", "KC_ENTER":"KC_ENT", "KC_ESCAPE":"KC_ESC", "KC_BSCAPE":"KC_BSPC", "KC_SPACE":"KC_SPC", "KC_MINUS":"KC_MINS",
        "KC_LBRACKET":"KC_LBRC", "KC_RBRACKET":"KC_RBRC", "KC_BSLASH":"KC_BSLS", "KC_NONUS_HASH":"KC_NUHS", "KC_NONUS_BSLASH":"KC_NUBS",
        "KC_INT1":"KC_RO", "KC_INT2":"KC_KANA", "KC_INT3":"KC_JYEN", "KC_INT4":"KC_HENK",  "KC_INT5":"KC_MHEN", "KC_SCOLON":"KC_SCLN",
        "KC_QUOTE":"KC_QUOT", "KC_GRAVE":"KC_GRV", "KC_COMMA":"KC_COMM", "KC_SLASH":"KC_SLSH", "KC_CAPSLOCK":"KC_CAPS", "KC_LCTRL":"KC_LCTL",
        "KC_LSHIFT":"KC_LSFT", "KC_LCMD":"KC_LGUI", "KC_LWIN":"KC_LGUI", "KC_RCTRL":"KC_RCTL", "KC_RSHIFT":"KC_RSFT", "KC_RCMD":"KC_RGUI", 
        "KC_RWIN":"KC_RGUI", "KC_LCAP":"KC_LOCKING_CAPS", "KC_LNUM":"KC_LOCKING_NUM", "KC_LSCR":"KC_LOCKING_SCROLL", "KC_PSCREEN":"KC_PSCR", 
        "KC_SCROLLLOCK":"KC_SLCK", "KC_PAUSE":"KC_PAUS", "KC_INSERT":"KC_INS", "KC_DELETE":"KC_DEL", "KC_DELT":"KC_DEL", "KC_PGDOWN":"KC_PGDN", 
        "KC_RIGHT":"KC_RGHT", 

        "KC_APPLICATION":"KC_APP", "KC_SYSTEM_POWER":"KC_PWR", "KC_SYSTEM_SLEEP":"KC_SLEP", "KC_SYSTEM_WAKE":"KC_WAKE", "KC_CALCULATOR":"KC_CALC", 
        "KC_MY_COMPUTER":"KC_MYCM", "KC_WWW_SEARCH":"KC_WSCH", "KC_WWW_HOME":"KC_WHOM", "KC_WWW_BACK":"KC_WBAK", "KC_WWW_FORWARD":"KC_WFWD", 
        "KC_WWW_STOP":"KC_WSTP", "KC_WWW_REFRESH":"KC_WREF", "KC_WWW_FAVORITES":"KC_WFAV", "KC_AUDIO_MUTE":"KC_MUTE", "KC_AUDIO_VOL_UP":"KC_VOLU", 
        "KC_AUDIO_VOL_DOWN":"KC_VOLD", "KC_MEDIA_NEXT_TRACK":"KC_MNXT", "KC_MEDIA_PREV_TRACK":"KC_MPRV", "KC_MEDIA_STOP":"KC_MSTP", 
        "KC_MEDIA_PLAY_PAUSE":"KC_MPLY", "KC_MEDIA_SELECT":"KC_MSEL", "KC_NUMLOCK":"KC_NLCK", "KC_KP_SLASH":"KC_PSLS", "KC_KP_ASTERISK":"KC_PAST", 
        "KC_KP_MINUS":"KC_PMNS", "KC_KP_PLUS":"KC_PPLS", "KC_KP_ENTER":"KC_PENT", "KC_KP_1":"KC_P1", "KC_KP_2":"KC_P2", "KC_KP_3":"KC_P3", 
        "KC_KP_4":"KC_P4", "KC_KP_5":"KC_P5", "KC_KP_6":"KC_P6", "KC_KP_7":"KC_P7", "KC_KP_8":"KC_P8", "KC_KP_9":"KC_P9", "KC_KP_0":"KC_P0", 
        "KC_KP_DOT":"KC_PDOT", "KC_KP_EQUAL":"KC_PEQL", "KC_KP_COMMA":"KC_PCMM", "KC_TRANSPARENT":"KC_TRNS", "_______":"KC_TRNS", "KC_MS_UP":"KC_MS_U", 
        "KC_MS_DOWN":"KC_MS_D", "KC_MS_LEFT":"KC_MS_L", "KC_MS_RIGHT":"KC_MS_R", "KC_MS_BTN1":"KC_BTN1", "KC_MS_BTN2":"KC_BTN2", "KC_MS_BTN3":"KC_BTN3", 
        "KC_MS_BTN4":"KC_BTN4", "KC_MS_BTN5":"KC_BTN5", "KC_MS_WH_UP":"KC_WH_U", "KC_MS_WH_DOWN":"KC_WH_D", "KC_MS_WH_LEFT":"KC_WH_L", 
        "KC_MS_WH_RIGHT":"KC_WH_R", "KC_MS_ACCEL0":"KC_ACL0", "KC_MS_ACCEL1":"KC_ACL1", "KC_MS_ACCEL2":"KC_ACL2", "RGB_MODE_FORWARD":"RGB_MOD", 
        "KC_TILDE":"KC_TILD", "KC_EXCLAIM":"KC_EXLM", "KC_DOLLAR":"KC_DLR", "KC_PERCENT":"KC_PERC", "KC_CIRCUMFLEX":"KC_CIRC", "KC_AMPERSAND":"KC_AMPR", 
        "KC_ASTERISK":"KC_ASTR", "KC_LEFT_PAREN":"KC_LPRN", "KC_RIGHT_PAREN":"KC_RPRN", "KC_UNDERSCORE":"KC_UNDS", "KC_LEFT_CURLY_BRACE":"KC_LCBR", 
        "KC_RIGHT_CURLY_BRACE":"KC_RCBR", "KC_COLON":"KC_COLN", "KC_DOUBLE_QUOTE":"KC_DQUO",  "KC_DQT":"KC_DQUO", "KC_LEFT_ANGLE_BRACKET":"KC_LABK", 
        "KC_LT":"KC_LABK", "KC_RIGHT_ANGLE_BRACKET":"KC_RABK", "KC_QUESTION":"KC_QUES"
    }

    keyplus_yaml_template = '''
# Generated by Q2K Keymap parser by 2Cas (c) 2018
# https://github.com/2Cas/Q2K
# This file is released into the public domain as per the CC0 Public Domain
# Dedication (http://creativecommons.org/publicdomain/zero/1.0/)
---
<ERRORS>

report_mode: auto_nkro # options: auto_nkro, 6kro, nkro

devices:
  <KB_NAME>:
    id: 0
    layout: <LAYOUT_NAME>
    layout_offset: 0
    scan_mode:
      mode: col_row # options: col_row, pins, none
      rows: <ROWS>
      cols: <COLS>
      # Maps how keys are physically wired, to how they appear visually
      matrix_map: [
        <MATRIX_MAP>
      ]
      # Debounce settings
      debounce:
         debounce_time_press: 5
         debounce_time_release: 5
         trigger_time_press: 1
         trigger_time_release: 1
         parasitic_discharge_delay_idle: 10.0
         parasitic_discharge_delay_debouncing: 10.0
<KEYCODES>
layouts:
  <LAYOUT_NAME> :
    default_layer: 0
    layers: [
<LAYOUT>    ]
'''

    keyplus_yaml_keycode_template = '''
  <KEYCODE>:
    keycode: hold
    delay: 200
    tap_key: <TAP>
    hold_key: <HOLD>
'''


import q2k.core as core
import sys
import os
import threading
import traceback
import yaml

from tkinter import filedialog, Tk, E, W, LEFT, INSERT, NORMAL, ttk, Menu, Text, scrolledtext, messagebox, Button, Entry, StringVar, LabelFrame, Label, Frame
from tkinter.ttk import Combobox

class ConsoleText(Text):
    '''A Tkinter Text widget that provides a scrolling display of console
    stderr and stdout.'''

    class IORedirector(object):
        '''A general class for redirecting I/O to this Text gcwidget.'''
        def __init__(self,text_area):
            self.text_area = text_area

    class StdoutRedirector(IORedirector):
        '''A class for redirecting stdout to this Text widget.'''
        def write(self,str):
            self.text_area.write(str,False)

    class StderrRedirector(IORedirector):
        '''A class for redirecting stderr to this Text widget.'''
        def write(self,str):
            self.text_area.write(str,True)

    def __init__(self, master=None, cnf={}, **kw):
        '''See the __init__ for Tkinter.Text for most of this stuff.'''

        Text.__init__(self, master, cnf, **kw)

        self.started = False
        self.write_lock = threading.Lock()

        self.tag_configure('STDOUT',background='black',foreground='white')
        self.tag_configure('STDERR',background='black',foreground='red')

        self.config(state=NORMAL)
        self.bind('<Key>',lambda e: 'break') #ignore all key presses

    def start(self):

        if self.started:
            return

        self.started = True

        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr

        stdout_redirector = ConsoleText.StdoutRedirector(self)
        stderr_redirector = ConsoleText.StderrRedirector(self)

        sys.stdout = stdout_redirector
        sys.stderr = stderr_redirector

    def stop(self):

        if not self.started:
            return

        self.started = False

        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr

    def write(self,val,is_stderr=False):

        self.write_lock.acquire()

        self.insert('end',val,'STDERR' if is_stderr else 'STDOUT')
        self.see('end')

        self.write_lock.release()

class Window():

    def __init__(self):
        # ───────────────────────────────────────────────────────────────────────────────────────────       
        # init Q2K tkinter GUI
        # ───────────────────────────────────────────────────────────────────────────────────────────
        self.init_q2k_gui()  
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Load Core Q2K
        # ───────────────────────────────────────────────────────────────────────────────────────────
        self.output.start() # Ensure output from loading q2k gets printed to console
        self.q2k_app = core.application('keyplus', is_gui=True)
        self.dirs = self.q2k_app.dirs
        self.output.stop() # Stop output printing
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Set dynamic lists and enter main loop
        # ───────────────────────────────────────────────────────────────────────────────────────────
        self.set_dirs()
        self.window.mainloop()

    def init_q2k_gui(self):
        self.window = Tk()
        #window.option_add('*Dialog.msg.font', 'Helvetica 12')
        self.window.resizable(width=False, height=False)
        self.window.option_add('*font', 'Roboto -11')
        self.window.title("Q2K Keymap Utility")
        self.window.geometry('380x560')

        menu = Menu(self.window)
        new_item = Menu(menu, tearoff=0)
        #new_item.add_command(label='Settings')
        #menu.add_cascade(label='File', menu=new_item)
        menu.add_command(label='About',command=self.show_about)
        self.window.config(menu=menu)

        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Tabs
        # ──────────────────────────────────────────────────────────────────────────────────────────

        tab_control = ttk.Notebook(self.window)
        tab1 = ttk.Frame(tab_control)
        #tab2 = ttk.Frame(tab_control)
        tab_control.add(tab1, text='Keyplus')
        #tab_control.add(tab2, text='KBfirmware')
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # KEYPLUS TAB
        # ───────────────────────────────────────────────────────────────────────────────────────────
        #   Frame 1 - Combo Boxes
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Label
        keyplus_lbl = Label(tab1, text= 'Convert to Keyplus YAML')
        keyplus_lbl.grid(column=0, row=0, sticky=W+E)
        # Settings
        top_frame = LabelFrame(tab1, text='', bd=0, padx=15, pady=2)
        top_frame.grid(sticky=W+E)

        # ───────────────────────────────────────────────────────────────────────────────────────────
        #       QMK Firmware Directory Entry
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Label
        qmk_lbl = Label(top_frame, text='QMK Firmware Directory:')
        qmk_lbl.grid(column=0, row=0, sticky=W)
        # Text Box
        self.qmk_dir = StringVar()
        qmk_dir = self.qmk_dir
        qmkdir_entry = Entry(top_frame, textvariable=qmk_dir, width=44)
        qmkdir_entry.grid(column=0, row=1, sticky=W+E)
        # Button
        btn = Button(top_frame, text="..", command=lambda:self.find_directory(qmk_dir))
        btn.grid(column=1, row=1)

        # ───────────────────────────────────────────────────────────────────────────────────────────
        #       QMK Firmware Directory Entry
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Label
        keyplus_lbl = Label(top_frame, text='Output Directory:')
        keyplus_lbl.grid(column=0, row=2, sticky=W)
        # Text Box
        self.keyplus_dir = StringVar()
        keyplus_dir = self.keyplus_dir
        keyplus_dir_entry = Entry(top_frame, textvariable=keyplus_dir, width=44)
        keyplus_dir_entry.grid(column=0, row=3, sticky=W+E)
        # Button
        btn = Button(top_frame, text="..", command=lambda:self.find_directory(keyplus_dir))
        btn.grid(column=1, row=3)
        # ───────────────────────────────────────────────────────────────────────────────────────────
        #   Frame 2 - Combo Boxes
        # ───────────────────────────────────────────────────────────────────────────────────────────
        kb_opts = LabelFrame(tab1, text='', bd=0, padx=15, pady=5, height=50)
        kb_opts.grid(sticky=E+W)
        # ───────────────────────────────────────────────────────────────────────────────────────────
        #       Dynamic Combo Boxes
        #          | Keyboard               |  Rev                 |  
        #          | Keymap                 |  Template            | 
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Labels
        kb_lbl = Label(kb_opts, text='Keyboard:')
        rev_lbl = Label(kb_opts, text='Rev:')
        keymap_lbl = Label(kb_opts, text='Keymap:')
        template_lbl = Label(kb_opts, text='Template:')
        # Combo Boxes
        self.template = Combobox(kb_opts, width=22, state='readonly')
        self.keymap = Combobox(kb_opts, width=22, state='readonly')
        self.rev = Combobox(kb_opts, width=22, state='readonly')
        self.rev.bind('<<ComboboxSelected>>', self.event_rev_selected) # Updating list triggers event rev_select
        self.kb = Combobox(kb_opts, width=22,  state='readonly')
        self.kb.bind('<<ComboboxSelected>>', self.event_kb_selected) # Updating list triggers event kb_select
        # Settings/Location
        kb_lbl.grid(sticky='w',column=0, row=0)
        rev_lbl.grid(sticky='w',column=1, row=0)
        keymap_lbl.grid(sticky='w',column=0, row=2)
        template_lbl.grid(sticky='w',column=1, row=2)
        self.kb.grid(column=0,row=1)
        self.rev.grid(column=1,row=1)
        self.keymap.grid(column=0,row=3)
        self.template.grid(column=1,row=3)
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Frame 3 - Buttons 
        # ───────────────────────────────────────────────────────────────────────────────────────────
        kb_opts2 = LabelFrame(tab1, text='', bd=0, padx=15, pady=5, height=50, width=100)
        kb_opts2.grid(sticky=E+W)
        # ───────────────────────────────────────────────────────────────────────────────────────────
        #     Buttons:  | Convert | Generate Keyboard List | Reset |
        # ───────────────────────────────────────────────────────────────────────────────────────────
        btn = Button(kb_opts2, text='Generate Keyboard List', command=lambda:self.btn_generate_lists())
        btn.grid(padx=5, row=0, column=1)
        btn = Button(kb_opts2, text='Reset', command=lambda:self.btn_reset())
        btn.grid(row=0, column=2)
        btn = Button(kb_opts2, text='Convert', command=lambda:self.btn_execute())
        btn.grid(row=0, column=0)
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Frame 4 - Output 
        # ───────────────────────────────────────────────────────────────────────────────────────────
        console = LabelFrame(tab1, text='', bd=0, padx=15, pady=5, height=50)
        console.grid(sticky=W+E)
        # ───────────────────────────────────────────────────────────────────────────────────────────
        #   Console
        # ───────────────────────────────────────────────────────────────────────────────────────────
        # Label
        con_lbl = Label(console, text='Console')
        con_lbl.grid(padx=5, row=0, column=0)
        # Text Box
        self.output = ConsoleText(console, height=27, width=56, bd=0, bg='black', fg='white', font='Consolas 8')
        self.output.grid(padx=5, row=1, column=0, sticky=W+E)

        # KBfirmware
        '''
        lbl2 = Label(tab2, text= 'Convert to KBfirmware JSON')
        lbl2.grid(column=0, row=0)

        frame1_qmk_dir = LabelFrame(tab2, text='  QMK Firmware Dir:  ', bd=0, padx=15, pady=2)
        frame1_qmk_dir.grid(sticky=E+W)

        qmkdir_entry = Entry(frame1_qmk_dir, textvariable=qmkdir, width=32)
        qmkdir_entry.grid(column=0, row=0, sticky=E+W)
        dir_btn = Button(frame1_qmk_dir, text="..", command=lambda:find_directory(qmkdir))
        dir_btn.grid(column=1, row=0)

        frame2_kbf_dir = LabelFrame(tab2, text='  Output Dir:  ', bd=0, padx=15, pady=2)
        frame2_kbf_dir.grid(sticky=E+W)

        kbfdir = StringVar()
        kbfdir_entry = Entry(frame2_kbf_dir, textvariable=kbfdir, width=32)
        kbfdir_entry.grid(column=0, row=0, sticky=E+W)
        dir_btn = Button(frame2_kbf_dir, text="..", command=lambda:find_directory(kbfdir))
        dir_btn.grid(column=1, row=0)
        '''

        tab_control.pack(expand=1, fill='both')

    def set_dirs(self):

        self.qmk_dir.set(self.dirs['QMK dir'])
        self.keyplus_dir.set(self.dirs['Keyplus YAML output'])

        kb_list = self.q2k_app.keyboard_list()
        kb_list.sort()
        self.kb['values'] = kb_list

    def find_directory(self, string):
        folder_selected = filedialog.askdirectory()
        string.set(folder_selected+'/')

    def show_about(self):
        messagebox.showinfo('About', 'Q2K Keymap Utility\nv '+core.defaults.version)


    def btn_generate_lists(self):
        self.output.start()

        self.dirs['QMK dir'] = self.qmk_dir.get()
        self.dirs['Keyplus YAML output'] = self.keyplus_dir.get()

        with open(core.defaults.src+'pref.yaml', 'w') as f:
            f.write('# Q2K Folder Locations\n')
            yaml.dump(self.dirs, f, default_flow_style = False)

        self.q2k_app.refresh_dir()
        self.q2k_app.refresh_cache()
        kb_list = self.q2k_app.keyboard_list()
        kb_list.sort()
        self.kb['values'] = kb_list
        self.kb.set('')
        self.rev.set('')
        self.keymap.set('')
        self.template.set('')

    def btn_reset(self):
        self.output.start()

        self.q2k_app.reset_dir_defaults()
        self.q2k_app.clear_cache()
        self.qmk_dir.set(core.defaults.qmk)
        self.keyplus_dir.set(core.defaults.keyp)
        self.kb.set('')
        self.rev.set('')
        self.keymap.set('')
        self.template.set('')
        self.btn_generate_lists()
        self.output.stop()

    def btn_execute(self):
        kb_n = self.kb.get()
        rev_n = self.rev.get()
        km_n = self.keymap.get()
        temp_n = self.template.get()

        try:
            self.output.start()
            self.q2k_app.set_kb(keyboard=kb_n, rev=rev_n, keymap=km_n, template=temp_n)
            self.q2k_app.execute()
            self.output.stop()

        except RuntimeError as e:
            #print(traceback.format_exc(), file=sys.stderr)
            print(str(e), file=sys.stderr)
            self.output.stop()

        except RuntimeWarning as e:
            print(str(e), file=sys.stderr)
            self.output.stop()

    def event_rev_selected(self, event=None):

        keymap_list = self.q2k_app.keymap_list(self.kb.get(), self.rev.get())
        template_list = self.q2k_app.template_list(self.kb.get(), self.rev.get())
        if keymap_list:
            keymap_list.sort()

        self.keymap['values'] = keymap_list
        self.template['values'] = template_list
        if keymap_list and 'default' in keymap_list:
            self.keymap.set('default')
        else:
            self.keymap.set('')
        if template_list:
            self.template.set(template_list[0])
        else:
            self.template.set('')

    def event_kb_selected(self, event):

        rev_list = self.q2k_app.rev_list(self.kb.get())
        rev_list.sort(reverse=True)

        if rev_list:

            self.rev.set(rev_list[0])
            self.rev['values']= rev_list
            self.event_rev_selected()

        else:

            keymap_list = self.q2k_app.keymap_list(self.kb.get())
            template_list = self.q2k_app.template_list(self.kb.get())
            if keymap_list:
                keymap_list.sort()

            self.rev['values']= rev_list
            self.keymap['values'] = keymap_list
            self.template['values'] = template_list


            self.rev.set('')
            if keymap_list and 'default' in keymap_list:
                self.keymap.set('default')
            else:
                self.keymap.set('')
            if template_list:
                self.template.set(template_list[0])
            else:
                self.template.set('')

def main():
    window = Window()



from setuptools import setup, find_packages

setup(
    name='q2k',
    version='1.0.7.a1',
    description='Convert AVR C based QMK keymap and matrix files to YAML Keyplus format',
    url='https://github.com/2Cas/Q2K',
    author='2Cas',
    author_email='reasonstouninstall@tutanota.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development',
        'Topic :: Text Processing',
        'License :: OSI Approved :: MIT License',
    ],
    license='MIT',
    packages=find_packages(include=['q2k', 'q2k.*']),
    python_requires='>=3.0',
    install_requires=[
        'pyyaml', 'pyparsing', 'termcolor'
    ],
    entry_points = {
        'console_scripts': ['q2k-cli=q2k.core:q2keyplus', 'q2k=q2k.gui:main']  #, 'q2kb=q2k.core:q2kbfirmware']
    },
    keywords = ['keyboard', 'usb', 'hid', 'qmk', 'keyplus'],
    zip_safe=False,
    include_package_data=True,
)



import argparse
import glob
import pathlib
import platform
import errno
import os
import yaml
import sys
import re
import copy
import subprocess
import pyparsing as pp
import termcolor as tc
import enum as en
import tkinter as tk

from q2k.reference import ref
# ───────────────────────────────────────────────────────────────────────────────────────────
# Defaults/Constants for Q2K
# ───────────────────────────────────────────────────────────────────────────────────────────
class defaults:

    if getattr(sys, 'frozen', False):
        # If Frozen - Use os.getcwd()
        src = os.getcwd().replace('\\', '/')+'/'
    else:
        # If Live, use bundle_dir
        frozen = False
        src = os.path.dirname(os.path.abspath(__file__)).replace('\\', '/')+'/'

    version = '1.0.6.a1'
    # Directories
    libs = src+'lib/'                                # Local Libs - We want these to be in a single, FIXED directory. [Avoid using relative directories]
    cache = src+'.cache/cache_kb.yaml'               # Cache-  We want this to be in a single, FIXED directory. [Avoid using relative directories]
    qmk = ''                                         # QMK Directory - To be provided by user
    keyp = os.getcwd().replace('\\', '/')+'/q2k_out/keyplus/'           # Output. This can either be fixed in one location (defined by user) or be a relative directory (default)
    kbf = os.getcwd().replace('\\', '/')+'/q2k_out/kbfirmware/'
    # Lists
    qmk_nonstd_dir = ['handwired', 'converter', 'clueboard', 'lfkeyboards'] # Currently only lfkeyboards causes any issues, however it is good to be verbose here
    q2k_incompat = ['handwired/']
    mcu_compat = ['atmega32u4', 'atmega32u2']                               # Compatible MCU types
    kbf_mcu_compat = ['atmega32u4', 'atmega32u2']

# ───────────────────────────────────────────────────────────────────────────────────────────
# Console Output
# ───────────────────────────────────────────────────────────────────────────────────────────
class _console:

    def __init__(self, gui):
        self.gui = gui
        self.errors = []

    def error(self, info):
        if self.gui:
            msg = ''
            for info, line in enumerate(info):
                msg += line+' \n'
                if not info:
                    warning=line
                    self.errors.append(line)
                    print('❌ ERROR: ' +line)
                else:
                    self.errors.append(line)
                    print('• '+line)

            tk.messagebox.showerror('Error', msg)
            raise RuntimeError(warning)
        else:
            error_msg = tc.colored('❌ ERROR:', 'red', attrs=['reverse', 'bold'])
            e_bullet = tc.colored('•', 'red', attrs=['bold'])
            
            for info, line in enumerate(info):
                if not info:
                    self.errors.append(line)
                    line = tc.colored(line, 'red')
                    print(error_msg + ' ' +line)
                else:
                    self.errors.append(line)
                    print(e_bullet + ' '+line)
            exit()

    def bad_kc(self, kc_type, code):
        if self.gui:
            message = 'Invalid '+kc_type+': '+code
            bad_kc_msg = '❌ Invalid '+kc_type+':'
            print(bad_kc_msg+' '+code)
            self.errors.append(message)
        else:
            message = 'Invalid '+kc_type+': '+code
            bad_kc_msg = tc.colored('❌ Invalid '+kc_type+':', 'cyan')
            print(bad_kc_msg+' '+code)
            self.errors.append(message)

    def warning(self, info, pause=False):
    
        if self.gui:
            msg = ''
            for info, line in enumerate(info):
                msg += line+' \n'
                if not info:
                    warning=line
                    self.errors.append(line)
                    print('▲ WARNING: ' +line)
                else: 
                    self.errors.append(line)
                    print('• '+line)
            if pause:
                if not tk.messagebox.askyesno('Warning', msg+'\nContinue?'):
                    raise RuntimeWarning(warning)

        else: 
            warning_msg = tc.colored('▲ WARNING:', 'yellow', attrs=['bold'])
            w_bullet = tc.colored('•', 'yellow', attrs=['bold'])
            
            for info, line in enumerate(info):
                if not info:
                    self.errors.append(line)
                    line = tc.colored(line, 'yellow')
                    print(warning_msg + ' ' +line)
                else:
                    self.errors.append(line)
                    print(w_bullet + ' '+line)
            if pause:
                print(w_bullet+' Press [ENTER] to continue')
                input()


    def note(self, info):
        
        if self.gui:
            for info, line in enumerate(info):
                if not info:
                    print('✔ ' +line)
                else:
                    print('• '+line)
        else:
            note_msg = tc.colored('✔', 'green', attrs=['bold'])
            n_bullet = tc.colored('•', 'green', attrs=['bold'])
            
            for info, line in enumerate(info):
                if not info:
                    print(note_msg + ' ' +line)
                else:
                    print(n_bullet + ' '+line)

    def clear(self):

        self.errors = []
# ───────────────────────────────────────────────────────────────────────────────────────────
# Text parsing
# ───────────────────────────────────────────────────────────────────────────────────────────
class _parse_txt:

    def layout_headers(data):

        LPAREN, RPAREN, LBRAC, RBRAC, COMMA = map(pp.Suppress, "(){},")
        BSLASH = pp.Suppress(pp.Literal('\\'))

        define = pp.Suppress(pp.Literal('#define'))
        name = pp.Word(pp.alphanums+'_')
        macrovar = pp.Word(pp.alphanums+'_#')
        layout_row = pp.Group(pp.OneOrMore(macrovar + pp.Optional(COMMA)) + pp.ZeroOrMore(BSLASH))
        layout =  LPAREN + pp.ZeroOrMore(BSLASH) + pp.OneOrMore(layout_row) + RPAREN
        matrix_row = pp.ZeroOrMore(BSLASH) + pp.Optional(LBRAC) +  pp.ZeroOrMore(BSLASH) + pp.OneOrMore(macrovar + pp.Optional(COMMA)) + pp.ZeroOrMore(RBRAC) + pp.Optional(COMMA) + pp.ZeroOrMore(BSLASH)
        matrix = pp.ZeroOrMore(BSLASH) + LBRAC +  pp.ZeroOrMore(BSLASH) + pp.OneOrMore(matrix_row) +  pp.ZeroOrMore(BSLASH) + RBRAC +  pp.ZeroOrMore(BSLASH)

        header = define + name('name') + pp.ZeroOrMore(BSLASH) + layout('layout') + pp.ZeroOrMore(BSLASH) + matrix('array')
        header.ignore(pp.cStyleComment)

        hresults = list(header.scanString(data))
        names = []
        for tokens, start, end in (hresults):
            i = 0
            while tokens.name in names:
                i += 1
                if not tokens.name.endswith(')'):
                    tokens.name += '('+str(i)+')'
                else:
                    tokens.name = tokens.name[:-2]+str(i)+')'
            names.append(tokens.name)
                
        return hresults

    def config_headers(data):

        data = str(data)
        matrix_pins = []
        matrix_row_pins = []
        matrix_col_pins = []
        
        LBRAC, RBRAC, COMMA = map(pp.Suppress, "{},")
        define_rows = pp.Suppress(pp.Literal('#define MATRIX_ROW_PINS'))
        define_cols = pp.Suppress(pp.Literal('#define MATRIX_COL_PINS'))
        pincode = pp.Word(pp.alphanums) | pp.Word(pp.nums)

        array = LBRAC + pp.ZeroOrMore(pincode + pp.Optional(COMMA)) + RBRAC
        matrix_rows = define_rows + LBRAC + pp.ZeroOrMore(pincode + pp.Optional(COMMA)) + RBRAC
        matrix_cols = define_cols + LBRAC + pp.ZeroOrMore(pincode + pp.Optional(COMMA)) + RBRAC

        for tokens, start, stop in matrix_rows.scanString(data):
            matrix_row_pins = list(tokens)
        for tokens, start, stop in matrix_cols.scanString(data):
            matrix_col_pins = list(tokens)

        if matrix_row_pins and matrix_col_pins:
            matrix_pins.append(matrix_row_pins)
            matrix_pins.append(matrix_col_pins)

        return matrix_pins

    def rules_mk(data):

        data = str(data)
        EQUALS = pp.Suppress('=')
        mcu_tag = pp.Suppress(pp.Literal('MCU'))
        mcu_type = pp.Word(pp.alphanums+'_')

        mcu = mcu_tag + EQUALS + mcu_type('mcu')
        mcu.ignore('#'+pp.restOfLine)

        return mcu.scanString(data)

    def keymaps(data):

        data = str(data)
        LBRAC, RBRAC,EQ, COMMA = map(pp.Suppress,"{}=,")

        integer = pp.Word(pp.nums)
        keycode = pp.Word(pp.alphanums+'_'+'('+')')
        layer_string = pp.Word(pp.printables) | pp.Word(pp.nums)

        keycode_list = pp.Group(pp.ZeroOrMore(keycode + pp.Optional(COMMA)))
        row = LBRAC + keycode_list + RBRAC

        layern = layer_string('layer_name') + EQ
        km_layer_data = LBRAC + pp.OneOrMore(row + pp.Optional(COMMA)) + RBRAC
        km_layer = pp.Optional(layern) + km_layer_data('layer') + pp.Optional(COMMA)
        km_layer.ignore(pp.cppStyleComment+pp.pythonStyleComment)

        return km_layer.scanString(data)

    def keymap_functions(data):

        data = str(data)
        LSQBRAC, RSQBRAC, EQ, COMMA = map(pp.Suppress,"[]=,")
        words = pp.Word(pp.alphanums+'_')
        function_header = pp.Literal('fn_actions[] = {')

        func_index = LSQBRAC + pp.Word(pp.alphanums+'_') + RSQBRAC + EQ
        func_name = words
        func_params = pp.Combine( '(' + pp.ZeroOrMore(words + pp.Optional(',')) + ')', adjacent=False)
        func = pp.Group(pp.Optional(func_index) + pp.Combine(func_name + func_params, adjacent=False))
        func_list = pp.Group(pp.ZeroOrMore( func + pp.Optional(COMMA) ) )
        EOF = pp.Literal('}')

        function = pp.Suppress(function_header) + func_list('function') + pp.Suppress(EOF)
        function.ignore(pp.cppStyleComment)

        results = list(function.scanString(data))

        return results
# ───────────────────────────────────────────────────────────────────────────────────────────
# Pre-processing
# ───────────────────────────────────────────────────────────────────────────────────────────
class _cpp:

    def __init__(self, kbo, dirs, console):
        self.__kb = kbo
        self.__dirs = dirs
        self.__console = console

    def __preproc(self, kblibs, arg_list, DEBUG=False):
        # Setting up -I and custom define options
        qdir = self.__dirs['QMK dir'] + 'keyboards/'
        kb = self.__kb.name
        if platform.system() == 'Linux':
            cc = ['avr-gcc', '-E']
        elif platform.system() == 'Windows':
            avr_gcc = defaults.src + 'avr-gcc/bin/avr-gcc.exe'
            cc = [avr_gcc, '-E']
        kbdefine = 'KEYBOARD_'+'_'.join(kblibs)
        QMK_KEYBOARD_H = 'QMK_KEYBOARD_H=\"'+kb+'.h\"'
        libs = ['-D', kbdefine, '-D', QMK_KEYBOARD_H, '-I'+self.__dirs['Local libs']]
        path = qdir
        for kbl in kblibs:
            kbl += '/'
            path += kbl
            libs.append('-I'+path)
        argv = cc + libs + arg_list
        if DEBUG: print(' '.join(argv))
        try:
            output = subprocess.check_output(argv)
            return output
        except subprocess.CalledProcessError as e:
            err_num = e.returncode
            if err_num == 1:
                print(err_num)
                output = e.output
                return output
            else:
                self.console.warning(['Potentially catastrophic segfault related compilation error'])
                return
    
    def preproc_header(self, path):

        kblibs = list(self.__kb.libs)
        if self.__kb.build_rev:
            kblibs.append(self.__kb.build_rev)

        arg = [ '-D', 'QMK_KEYBOARD_CONFIG_H=\"config_common.h\"', '-dD', path ]
        if os.path.isfile(path):
            output = str(self.__preproc(kblibs, arg)).replace('\\n', ' ')
            return output

    def preproc_keymap(self):

        qdir = self.__dirs['QMK dir'] +'keyboards/'
        kb = self.__kb.name
        km = self.__kb.build_keymap+'/keymap.c'
        kblibs = list(self.__kb.libs)
        rev = self.__kb.build_rev
        if rev != '':
            kblibs.append(rev)

        # KEYMAP
        keym = qdir+kb+'/keymaps/'+km
        if rev != '' and not os.path.isfile(keym):
            keym = qdir+kb+'/'+rev+'/keymaps/'+km

        # OUTPUT
        argkm = [keym]
        if os.path.isfile(keym):
            output = str(self.__preproc(kblibs, argkm)).replace('\\n', ' ')
            return output
        else:
            self.__console.error(['Keymap cannot be read by preprocessor', 'Failed to parse keymap file'])
# ───────────────────────────────────────────────────────────────────────────────────────────
# KB Information
# ───────────────────────────────────────────────────────────────────────────────────────────
class kb_info:

    def __init__(self, n=''):

        self.name = n                       # Name of keyboard
        self.libs = []                      # Possible QMK lib folders (partly depreciated)
        self.rev_list = []                  # List of rev obj names
        self.rev_info = []                  # A list of rev_info objects

    def init_build(self):

        self.build_rev = ''               # What revision to build with
        self.build_keymap = ''              # What keymap to build with
        self.build_template = ''            # What layout to build with

        #self.build_m_row_pins = []          # Row pins
        #self.build_m_col_pins = []          # Column 

    def add_rev_list(self, rev, flag=False):
        if not flag:
            self.rev_list.append(rev)
        revo = rev_info(rev, flag)
        self.rev_info.append(revo)

    def get_rev_info(self, rev):
        for r in self.rev_info:
            if r.name == rev:
               return r
            if r.name == 'n/a':
               return r
#───────────────────────────────────────────────────────────────────────────────────────────
# KB/revision Information
# ───────────────────────────────────────────────────────────────────────────────────────────
class rev_info:

    def __init__ (self, n='', flag=False):

        self.name = n                      # Name of Revision
        self.keymap_list = []              # List of layout NAMES
        self.template_list = []            # List of template NAMES
        self.template_loc = ''             # Location of <keyboard>.h
        self.is_default = flag             # Does keyboard have revisions? (or just default)

    def init_build(self):
        # TODO: Possibly move ALL build information to kb_info? Or keep it like it is?

        self.build_mcu_list = []                 # MCU list
        self.build_m_row_pins = []               # Row pins
        self.build_m_col_pins = []               # Column  pins

        self.build_layout = []                   # list of keycode_layers objects (which form layout) -> Final yaml/json output comes from here
        self.build_templates = []                # list of layout_template objects

# ───────────────────────────────────────────────────────────────────────────────────────────
# LAYOUT template info
# A class for linking matrix mapping in <keyboard>.h with (preprocessed) keymap.c keycode layers.
# ───────────────────────────────────────────────────────────────────────────────────────────
class layout_template:

    def __init__(self, n=''):
        self.name = n                      # Name of template : e.g. LAYOUT, KEYMAP, LAYOUT_66_ANSI
        self.layout = []                   # List of layout rows
        self.array = []                    # Array holding index values (to be merged with keycode_layers)

    def convert_template_index(self, console=_console(False), template_list=[]):

        for i, row in enumerate(self.layout):
            for j, col in enumerate(row):
                try:
                    self.layout[i][j] = self.array.index(col)
                except ValueError:
                    console.warning([self.name+' - missing macro variable ['+str(col)+']'])
                    # Try to recover using a different keycode array
                    if template_list:
                        for t in template_list:                    
                            array2 = t.array
                            if len(array2) != len(self.array) or self.name == t.name:
                                continue
                            try:
                                index = array2.index(col)
                                self.layout[i][j] = index
                                self.array[index] = col                         
                                console.warning(['Macro variable recovery succeeded'])
                                break
                            except ValueError:
                                continue
                    if self.layout[i][j] == col:
                        self.layout[i][j] = -1
                        console.warning(['Array key recovery failed', 'Will assume this corresponds to KC_NO'])
# ───────────────────────────────────────────────────────────────────────────────────────────
# KEYCODE layers
# A class for storing keycode layers
# ───────────────────────────────────────────────────────────────────────────────────────────
class keycode_layer:

    def __init__(self, n=''):

        self.name = n                      # Name of current layer
        self.keymap = []                   # List of lists representing Raw [QMK] KEYMAP layers (keycode layers)
        self.layout = []                   # List of lists output for converted [keyplus] keycode layers/layout
        self.matrix_map = []               # Matrix mapping in [keyplus] format
        self.matrix_cols = 0

    def convert_keyplus_keymap(self, layer_names, functions, console):

        for i, kc in enumerate(self.keymap):
            if kc.endswith(')') and '(' in kc:
                self.keymap[i] = self.__func(kc, layer_names, functions, console)
            else:
                self.keymap[i] = self.__keycode(kc, functions, console)

    def convert_keyplus_matrix(self, col_limit):

        matrix = copy.deepcopy(self.matrix_map)
        for i, row in enumerate(self.matrix_map):
            j = 0
            for col in row:
                int(col)
                if col != -1:
                    r = col // col_limit
                    c = col % col_limit
                    matrix[i][j] = 'r'+str(r)+'c'+str(c)
                    j += 1
                else:
                    del matrix[i][j]
                    del self.layout[i][j]
                    j = max(j-1, 0)
        self.matrix_map = matrix

    def __func(self, qmk_func, layer_names, functions, console):

        # ───────────────────────────────────────────────────────────────────────────────
        # OSM Functions
        # ───────────────────────────────────────────────────────────────────────────────
        if qmk_func in ref.keyp_func.keys():
            keyp_func = ref.keyp_func[qmk_func]
        else:
            # Break up into [qfunc]([func_target])
            brac_index = qmk_func.index('(')+1
            qfunc = qmk_func[:brac_index]
            func_target = qmk_func[brac_index:-1]
            # ───────────────────────────────────────────────────────────────────────────────
            # Layer Switching Functions
            # ───────────────────────────────────────────────────────────────────────────────
            if qfunc in ref.keyp_layer_func.keys() and func_target in layer_names:
                layer = str(layer_names.index(func_target))
                keyp_func = ref.keyp_layer_func[qfunc]+layer
            # ───────────────────────────────────────────────────────────────────────────────
            # Mod/Multi-Mod Functions
            # ───────────────────────────────────────────────────────────────────────────────
            elif qfunc in ref.keyp_func.keys() and func_target in ref.keyp_kc.keys():
                #keycode = self.__keycode(func_target, functions, console)
                keycode = ref.keyp_kc[func_target]
                if keycode == "\"\'\" ":
                    keycode = 'quot'
                keycode = keycode.replace("'", '')
                keyp_func = "'"+(ref.keyp_func[qfunc]+'-'+keycode).replace(' ','')+"'"
            # ───────────────────────────────────────────────────────────────────────────────
            # For Layer-Tap Keys e.g. LT(1, KC_SPACE) -> Space-FN
            # ───────────────────────────────────────────────────────────────────────────────
            elif qfunc in ref.keyp_tap_layer.keys():
                split = func_target.split(',', 1)
                if len(split) != 2:
                    console.bad_kc('FN','['+qmk_func+'] - set to trns')
                    keyp_func = 'trns'
                    return keyp_func
                else:
                    layer_target = split[0].replace(' ','')
                    keycode = split[1].replace(' ','')

                if layer_target not in layer_names:
                    console.bad_kc('FN','['+qmk_func+'] - set to trns')
                    keyp_func = 'trns'
                    return keyp_func
                else:
                    layer = str(layer_names.index(layer_target))
                    hold = ref.keyp_tap_layer[qfunc]+layer

                    if keycode not in ref.keyp_kc.keys() and keycode not in ref.qmk_legacy_mod.keys():
                        console.bad_kc('FN','['+qmk_func+'] - set to trns')
                        keyp_func = 'trns'
                        return keyp_func

                    elif keycode in ref.keyp_kc.keys():
                        tap = ref.keyp_kc[keycode]
                        if tap == "\"\'\" ":
                            tap = 'quot'
                        tap = tap.replace("'", '')
                        keyp_func = "'"+(tap +'>'+hold).replace(' ','')+"'"

                    elif keycode in ref.qmk_legacy_mod.keys():
                        tap = ref.qmk_legacy_mod[keycode]+'-none'
                        if tap == "\"\'\" ":
                            tap = 'quot'
                        tap = tap.replace("'", '')
                        keyp_func = "'"+(tap +'>'+hold).replace(' ','')+"'"
            # ───────────────────────────────────────────────────────────────────────────────
            # Modifier Tap e.g. RSFT_T(KC_UP) -> Up when Tapped, RShift when Held
            # ───────────────────────────────────────────────────────────────────────────────
            elif qfunc in ref.keyp_tap_mod.keys() and func_target in ref.keyp_kc.keys():
                hold = ref.keyp_tap_mod[qfunc]+'-none'
                tap = ref.keyp_kc[func_target]
                if tap == "\"\'\" ":
                    tap = 'quot'
                tap = tap.replace("'", '')
                keyp_func = "'"+(tap+'>'+hold).replace(' ','')+"'"
            # ───────────────────────────────────────────────────────────────────────────────
            # Chained Quantum Functions [Legacy] - i.e. LCTL(LALT(KC_DEL))
            # ───────────────────────────────────────────────────────────────────────────────
            elif qfunc in ref.keyp_func.keys() and ')' in func_target:
                target = func_target
                keyp_func = ref.keyp_func[qfunc]
                while ')' in target:
                    brac_index = target.index('(')+1
                    func = target[:brac_index]
                    target = target[brac_index:-1]
                    if func in ref.keyp_func.keys():
                        keyp_func += ref.keyp_func[func]

                if target in ref.keyp_kc.keys():
                    #final_kc = self.__keycode(target, functions, console)
                    final_kc = ref.keyp_kc[target]
                    if final_kc == "\"\'\" ":
                        final_kc = 'quot'
                    final_kc = final_kc.replace("'", '')
                    keyp_func = "'"+ (keyp_func+'-'+final_kc).replace(' ','')+"'"
                else:
                    console.bad_kc('FN','['+qmk_func+'] - set to trns')
                    keyp_func = 'trns'
            # ───────────────────────────────────────────────────────────────────────────────
            # Legacy TMK-style QMK Functions
            # ───────────────────────────────────────────────────────────────────────────────
            elif qfunc in ref.qmk_legacy_func:
                # Check for blank function list
                if not functions:
                    console.bad_kc('FN','['+qmk_func+'] - set to trns')
                    keyp_func = 'trns'
                else:
                    func_action = functions[int(func_target)]
                    # Check for blank function
                    if not func_action:
                        console.bad_kc('FN','['+qmk_func+'] - set to trns')
                        keyp_func = 'trns'
                    else:
                        keyp_func = func_action
            # ───────────────────────────────────────────────────────────────────────────────
            # Legacy QMK Modkey format e.g. MT(MOD_LCTL, KC_Z) -> Ctrl+Z
            # ───────────────────────────────────────────────────────────────────────────────
            elif qfunc == 'MT(':

                split = func_target.split(',', 1)
                if len(split) != 2:
                    console.bad_kc('FN','['+qmk_func+'] - set to trns')
                    keyp_func = 'trns'
                    return keyp_func
                else:
                    modifier = split[0]
                    keycode = split[1]

                if modifier in ref.qmk_legacy_mod.keys() and keycode in ref.keyp_kc.keys():
                    mod = ref.qmk_legacy_mod[modifier]
                    kc = ref.keyp_kc[keycode]
                    if kc == "\"\'\" ":
                        kc = 'quot'
                    kc = kc.replace("'", '')
                    keyp_func = "'"+(mod+'-'+kc).replace(' ','')+"'"

            else:
                console.bad_kc('FN','['+qmk_func+'] - set to trns')
                keyp_func = 'trns'

        return keyp_func

    def __keycode(self, qmk_kc, functions, console):

        if qmk_kc in ref.keyp_kc.keys():

            keyp_kc = ref.keyp_kc[qmk_kc]

        elif qmk_kc[:5] in ref.qmk_legacy_func:
            # Check for blank function list
            if not functions:
                console.bad_kc('FN','['+qmk_kc+'] - set to trns')
                keyp_kc = 'trns'
            else:
                func_index = int(qmk_kc[5:])
                # Check array out of bounds
                if func_index < len(functions):
                    func_action = functions[func_index]
                    # Check for blank function
                    if func_action:
                        keyp_kc = func_action
                    else:
                        console.bad_kc('FN','['+qmk_kc+'] - set to trns')
                        keyp_kc = 'trns'
                else:
                    console.bad_kc('FN','['+qmk_kc+'] - set to trns')
                    keyp_kc = 'trns'

        else:
            console.bad_kc('KC','['+qmk_kc+'] - set to trns')
            keyp_kc = 'trns'
        
        return keyp_kc
# ───────────────────────────────────────────────────────────────────────────────────────────
# Cached Lists and Dictionaries
# ───────────────────────────────────────────────────────────────────────────────────────────
class _cache:

    def __init__(self, dirs, console, f_cache=False):

        self.kbo_list = []

        self.__loc = dirs['Cache']
        self.__qmk = dirs['QMK dir']
        self.__console = console

        if not f_cache: 
            self.__find()
        else:
            self.__write()

    def __find(self):
        if os.path.isfile(self.__loc):
            try:
                with open(self.__loc, 'r') as f:
                    self.kbo_list = yaml.load(f)
                    self.__console.note(['Using cached list from '+self.__loc, '--cache to reset'])
            except:
                self.__console.warning(['Failed to load from '+self.__loc, 'Generating new cache_kb.yaml...'])
                self.__write()
        else:
                self.__write()

    def __write(self):

        templist = []
        keymaplist = []

        for fn in glob.glob(self.__qmk+'keyboards/**/rules.mk', recursive=True):
            fn = fn[:-9].replace('\\','/')
            templist.append(fn)

        for fn in glob.glob(self.__qmk+'keyboards/**/keymaps/**/keymap.c', recursive=True):
            fn = fn[:-9].replace('\\','/')
            if fn in templist:
                templist.remove(fn)
            fn = fn.replace(self.__qmk+'keyboards/', '', 1)
            keymaplist.append(fn)

        for child in templist:
            p_path = str(pathlib.Path(child).parent).replace('\\','/')
            p_name = p_path.replace(self.__qmk+'keyboards/', '', 1)

            if p_path not in templist and p_name not in defaults.qmk_nonstd_dir:
                name = child.replace(self.__qmk+'keyboards/', '', 1)
                # If in special dir list, then is part of the directory not keyboard
                if name not in defaults.qmk_nonstd_dir:
                   # This is a keyboard, not revision
                   kbo = kb_info(name)
                   kblibs = name.split('/')
                   kbo.libs = kblibs 
                   self.kbo_list.append(kbo)   
            elif p_name in defaults.qmk_nonstd_dir:
                name = child.replace(self.__qmk+'keyboards/', '', 1)
                # This is a keyboard, not revision
                kbo = kb_info(name)
                kblibs = name.split('/')
                kbo.libs = kblibs
                self.kbo_list.append(kbo)
            else:
                # This is a 'revision' of an existing keyboard
                rev = child.replace(p_path+'/', '')
                for kbo in self.kbo_list:
                    if kbo.name == p_name:
                        kbo.add_rev_list(rev)
                        break

        for kbo in self.kbo_list:
            if not kbo.rev_list:
                kbo.add_rev_list('n/a', True)
            self.__find_layout_names(kbo)

        for km_path in keymaplist:
            info_list = km_path.split('/keymaps/')
            kb_name = info_list[0]
            km_name = info_list[-1]
            namelist = kb_name.split('/')
            match = False
            prev = 'n/a'
            for i in range(0, len(namelist)):
                if i > 0: 
                    kb_name = '/'.join(namelist)

                for kbo in self.kbo_list:
                    if kbo.name == kb_name:
                        if prev in kbo.rev_list:
                            revo = kbo.get_rev_info(prev)
                            revo.keymap_list.append(km_name)
                        else:
                            for revo in kbo.rev_info:
                                revo.keymap_list.append(km_name)
                        match = True
                        break
                if match:
                    break
                prev = namelist.pop()

        if self.kbo_list:
            # Dump cache info to text file for faster processing in future
            self.__save_cache()
            self.__console.note(['New cache_kb.yaml successfully generated', 'Location: '+self.__loc])
        else:
            self.__console.warning(['No keyboard information found', 'Check QMK directory location in pref.yaml : '+self.__qmk])
        

    def __find_layout_names(self, kbo):
        
        rev_list = list(kbo.rev_list)
        if not rev_list:
            rev_list.append('n/a')
        
        for rev in rev_list:
            found = False
            revo = kbo.get_rev_info(rev)
            # qmk/keyboards/
            kblibs = list(kbo.libs)
            if rev != 'n/a':
                kblibs.append(rev)
            qdir = self.__qmk +'keyboards/'

            folders = []
            path = ''
            for kbl in kblibs:
                path += kbl+'/'
                folders.append(path)

            for kbl in reversed(folders):
                kbl_f = kbl.split('/')
                kb_h = kbl_f[-2]        
                kb_h += '.h'
                path = qdir+kbl+kb_h

                try:
                    with open(path, 'r', encoding='utf8') as f:
                        data = str(f.read())
                except FileNotFoundError:
                    #self.__console.warning(['Layout header not found in '+path, 'Trying a different path...'])
                    continue

                token_list = _parse_txt.layout_headers(data)
                #template_list = []
                for tokens, start, end in token_list:
                    revo.template_list.append(tokens.name)

                if revo.template_list:
                    #self.__console.note(['Layout header found @ '+path])
                    found = True
                    revo.template_loc = path
                    break

            if rev != 'n/a' and not found:
                self.__console.warning(['Layout templates not found for '+kbo.name+'/'+rev])
                revo.template_loc = 'n/a'
            elif not found:
                self.__console.warning(['Layout templates not found for '+kbo.name])
                revo.template_loc = 'n/a'

    def __save_cache(self):

        path = '/'.join(self.__loc.split('/')[:-1])
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno != errno.EEXIST and os.path.isdir(path):
                    raise
        try:
            with open(self.__loc, 'w') as f:
                yaml.dump(self.kbo_list, f)
        except:
            self.__console.error(['Failed to create '+self.__loc])

    def _clear_cache(self):
        if os.path.isfile(self.__loc):
            os.remove(self.__loc)

    def _keyboard_list(self,):
        kb_names = []
        for kbo in self.kbo_list:
            kb_names.append(kbo.name)
        return kb_names

    def _keymap_list(self, keyboard, rev=''):
        km_names = []
        for kbo in self.kbo_list:
            if kbo.name == keyboard:
                revo = kbo.get_rev_info(rev)
                if revo:
                    for km in revo.keymap_list:
                        km_names.append(km)
                    return km_names
                else: 
                    print_rev_list = ', '.join(self._rev_list(keyboard))
                    self.__console.error(['Revision # required - Valid Revisions: '+print_rev_list])

    def _rev_list(self, keyboard):

        for kbo in self.kbo_list:
            if kbo.name == keyboard:
                return kbo.rev_list

    def _template_list(self, keyboard, rev=''):
        tp_names = []
        for kbo in self.kbo_list:
            if kbo.name == keyboard:
                revo = kbo.get_rev_info(rev)
                if revo:
                    for tp in revo.template_list:
                        tp_names.append(tp)
                    return tp_names
                else:
                    print_rev_list = ', '.join(self._rev_list(keyboard))
                    self.__console.error(['Revision # required - Valid Revisions: '+print_rev_list])
# ───────────────────────────────────────────────────────────────────────────────────────────
# Q2K Application
# ───────────────────────────────────────────────────────────────────────────────────────────
class application:

    def __init__(self, app_type, is_gui=False):
        self.__output = en.Enum('output', 'keyplus kbfirmware')
        self.format = self.__output[app_type]
        self.is_gui = is_gui
        self.console = _console(is_gui)
        # self.dirs      # directories
        # self.build_kb  # kb_info for build
        # self.build_rev # rev_info for build
        # self.__args    # Cmd line arguments
        # self.__cache,  # kb_list cache
        if not is_gui:
            self.__read_args()                             # init self.__args
            self.__set_dirs()                              # Read pref.yaml (or create)
            self.__pop_cache_list()                        # Get cached kb info and put into list
            self.__check_args()
        else:
            self.__read_args()
            self.__set_dirs()                              # Skip reading cmd line (set kb with hook-in method)
            self.__pop_cache_list()                        # Get cached kb info and put into list                             

    def __read_args(self):
        # Read ARGV input from terminal
        parser = argparse.ArgumentParser(description='Convert AVR C based QMK keymap and matrix files to YAML Keyplus format')
        parser.add_argument('keyboard', metavar='KEYBOARD', nargs='?', default='', help='The name of the keyboard whose keymap you wish to convert')
        parser.add_argument('-m', '--keymap', metavar='KEYMAP', dest='keymap', default='default', help='The keymap folder to reference - default is /default/')
        parser.add_argument('-t', '--template', metavar='LAYOUT', dest='template', default='', help='The layout template to reference')
        parser.add_argument('-r', '--rev', metavar='ver',dest='rev', default='', help='Revision of layout - default is n/a')
        parser.add_argument('--cache', dest='clearcache', action='store_true', help='Clear cached data (cache_kb.yaml)')
        parser.add_argument('--reset', dest='clearpref', action = 'store_true', help='Restore preferences in pref.yaml to default')
        parser.add_argument('--debug', dest='debug', action='store_true', help='See debugging information')
        parser.add_argument('-l', '-L', '--list', dest='listkeyb', action='store_true',help='List all valid KEYBOARD inputs')
        parser.add_argument('-M', '--keymaps', dest='listkeym', action='store_true',help='List all valid KEYMAPS for the current keyboard')
        parser.add_argument('-T', '--templatelist', dest='listkeyt', action='store_true',help='List all valid LAYOUTS for the current keyboard')
        parser.add_argument('-R', '--revlist', dest='listkeyr', action='store_true',help='List all valid REVISIONS for the current keyboard')
        parser.add_argument('-S', '--search', metavar='string', dest='searchkeyb', help='Search valid KEYBOARD inputs')
        self.__args = parser.parse_args()

    def __set_dirs(self):

        if self.__args.clearpref:
            self.__generate_dirs()
        else:
            try:
                with open(defaults.src+'pref.yaml', 'r') as f:
                    self.dirs = yaml.load(f)
                    self.console.note(['─────────────────────────────────────────────────────────────────', 'Using preferences from '+defaults.src+'pref.yaml', '--reset to reset to defaults'])

            except FileNotFoundError:
                self.__generate_dirs()

            except AttributeError:
                self.__generate_dirs()

    def __generate_dirs(self):

        dirs = {
            'QMK dir' : defaults.qmk,
            'Keyplus YAML output' : defaults.keyp,
            'Kbfirmware JSON output' : defaults.kbf,
            'Local libs' : defaults.libs,
            'Cache' : defaults.cache,            
        }
        self.dirs = dirs

        with open(defaults.src+'pref.yaml', 'w') as f:
            f.write('# Q2K Folder Locations\n')
            yaml.dump(dirs, f, default_flow_style = False)
            self.console.note(['─────────────────────────────────────────────────────────────────', 'New pref.yaml generated'])

    def __pop_cache_list(self):

        self.__cache = _cache(self.dirs, self.console, self.__args.clearcache)

    def __check_args(self):

        if self.__args.listkeyb and not self.is_gui:
            print_kb_list = '[ '+', '.join(self.keyboard_list())+' ]'
            self.console.note(['Listing keyboards...', print_kb_list])
            exit()
        elif self.__args.listkeyr and not self.is_gui:
            print_rev_list = '[ '+', '.join(self.rev_list(self.__args.keyboard))+' ]'
            self.console.note(['Listing revisions for '+self.__args.keyboard+'...', print_rev_list])
            exit()
        elif self.__args.listkeym and not self.is_gui:
            print_km_list = '[ '+', '.join(self.keymap_list(self.__args.keyboard, self.__args.rev))+' ]'
            if self.__args.rev == '':
                self.console.note(['Listing keymaps for '+self.__args.keyboard+'...', print_km_list])
            else:
                self.console.note(['Listing keymaps for '+self.__args.keyboard+'/'+self.__args.rev+'...', print_km_list])
            exit()
        elif self.__args.listkeyt and not self.is_gui:
            print_temp_list = '[ '+', '.join(self.template_list(self.__args.keyboard, self.__args.rev))+' ]'
            if self.__args.rev == '':
                self.console.note(['Listing layout templates for '+self.__args.keyboard+'...', print_temp_list])
            else:
                self.console.note(['Listing layout templates for '+self.__args.keyboard+'/'+self.__args.rev+'...', print_temp_list])
            exit()
        elif self.__args.searchkeyb and not self.is_gui:
            print_search_list = '[ '+', '.join(self.search_keyboard_list(self.__args.searchkeyb))+' ]'
            self.console.note(['Searching for '+self.__args.searchkeyb+'...', print_search_list])
            exit()

        self.set_kb(self.__args.keyboard, self.__args.rev, self.__args.keymap, self.__args.template)

    # Intended hook-in for GUIs
    def set_kb(self, keyboard='', rev='', keymap='', template=''):

        kb_list = self.__cache.kbo_list
        build_kbo = ''

        if not keyboard:
            print_kb_list = ', '.join(self.keyboard_list())
            self.console.error(['No keyboard name given', 'Valid Names: '+print_kb_list])

        for kbo in kb_list:
            if kbo.name == keyboard:
                build_kbo = kbo
                break

        if build_kbo:
            # Check Revision
            # Case 1: Have Revisions
            if len(build_kbo.rev_list) > 0 and rev not in build_kbo.rev_list:
                print_rev_list = ', '.join(self.rev_list(keyboard))
                self.console.error(['Invalid Revision - '+rev, 'Valid Revisions: '+print_rev_list])
            # Case 2: No Revisions
            elif len(build_kbo.rev_list) < 1 and rev != '':
                self.console.error(['Invalid Revision - '+rev, 'Valid Revisions: None'])

            build_revo = build_kbo.get_rev_info(rev)
            # Check Layout
            # Case 1: Have Layout Templates
            if len(build_revo.template_list) > 0 and template not in build_revo.template_list:
                print_temp_list = ', '.join(self.template_list(keyboard, rev))
                self.console.error(['Invalid Template - '+template, 'Valid Layouts: '+print_temp_list])
            # Case 2: No Layout Templates
            elif len(build_revo.template_list) < 1 and template != '':
                self.console.error(['Invalid Template - '+template, 'Valid Layouts: None'])

            # Check Keymap
            if keymap not in build_revo.keymap_list:
                print_km_list = ', '.join(self.keymap_list(keyboard))
                self.console.warning(['Invalid Keymap - '+template, 'Valid Keymaps: '+print_km_list])

            build_kbo.init_build()    
            build_revo.init_build()
            build_kbo.build_keymap = keymap
            build_kbo.build_template = template
            build_kbo.build_rev = rev
            self.build_kb = build_kbo
            self.build_rev = build_revo
            if rev:
                self.console.note(['─────────────────────────────────────────────────────────────────','Building '+keyboard+'/'+rev+':'+keymap+':'+template, '─────────────────────────────────────────────────────────────────'])
            else:
                self.console.note(['─────────────────────────────────────────────────────────────────','Building '+keyboard+':'+keymap+':'+template, '─────────────────────────────────────────────────────────────────'])
        else:
            print_kb_list = ', '.join(self.keyboard_list())
            self.console.error(['Invalid Keyboard Name - '+keyboard, 'Valid Names: '+print_kb_list])

    def refresh_dir(self):
        self.__set_dirs()

    def reset_dir_defaults(self):
        self.__generate_dirs()

    def refresh_cache(self):
        self.__cache = _cache(self.dirs, self.console, True)

    def clear_cache(self):
        self.__cache._clear_cache()

    def keyboard_list(self):
        return self.__cache._keyboard_list()

    def rev_list(self, keyboard):
        return self.__cache._rev_list(keyboard)

    def template_list(self, keyboard, rev=''):
        return self.__cache._template_list(keyboard, rev)

    def keymap_list(self, keyboard, rev=''):
        return self.__cache._keymap_list(keyboard, rev)

    def search_keyboard_list(self, string):

        kb_names = self.__cache._keyboard_list()
        results = []
        for kb in kb_names:
            if kb.find(string) != -1:
                results.append(kb)
        return results

    def execute(self):

        self.__cpp = _cpp(self.build_kb, self.dirs, self.console)

        self.console.clear()            # Clear console
        self.__check_mcu()              # Check for MCU and Matrix Pins
        self.__get_config_header()
        self.__get_keycodes()           # Init Layout + Templates
        self.__get_templates() 
        self.__merge_layout_template()  # Process Layout + Templates
        self.__convert_matrix_map()
        self.__create_output()          # Pipe output to yaml/json
        self.console.clear()            # Clear console

    def __check_mcu(self):
        self.__get_mcu()
        kb_n = self.build_kb.name
        revo = self.build_rev
        rev_n = self.build_kb.build_rev

        for mcu in revo.build_mcu_list:
            if mcu in defaults.mcu_compat:
                continue
            else:
                self.console.warning(
                    ['Possible MCU incompatability detected', 
                    mcu+' in '+kb_n+'/'+rev_n+'rules.mk', 
                    'Currently, keyplus supports only boards with the following microcontrollers:',
                    ', '.join(defaults.mcu_compat),
                    'If your board has a MCU on this list then ignore this warning as a false positive',
                    'Else layout files produced will not work with keyplus until your board\'s mcu is supported'], True)
                return

        self.console.note(['No MCU incompatability detected'])

    def __get_mcu(self):

        rev = self.build_rev.name
        revo = self.build_rev
        kblibs = list(self.build_kb.libs)
        if rev != 'n/a':
            kblibs.append(rev)

        qdir = self.dirs['QMK dir'] +'keyboards/'
        folders = []
        path = ''

        for kbl in kblibs:
            path += kbl+'/'
            folders.append(path)

        for kbl in reversed(folders):
            rules_mk = 'rules.mk'

            path = qdir+kbl+rules_mk
            mcu_list = []
            try:
                with open(path, 'r') as f:
                    data = str(f.read())
            except FileNotFoundError:
                self.console.warning(['Rules.mk not found in '+path, 'Trying a different path...'])
                return

            token_list = _parse_txt.rules_mk(data)
            for tokens, start, end in token_list:
                mcu_list.append(tokens.mcu)

            if mcu_list:
                revo.build_mcu_list = mcu_list
                return
            else:
                continue

        self.__console.warning(['Rules.mk not found for '+kbo.get_name()])
        return

    def __get_config_header(self):

        rev = self.build_rev.name
        revo = self.build_rev
        kblibs = list(self.build_kb.libs)
        if rev != 'n/a':
            kblibs.append(rev)

        qdir = self.dirs['QMK dir'] +'keyboards/'
        folders = []
        path = ''

        for kbl in kblibs:
            path += kbl+'/'
            folders.append(path)

        for kbl in reversed(folders):
            config_h = 'config.h'

            path = qdir+kbl+config_h
            data = self.__cpp.preproc_header(path)

            matrix_pins = _parse_txt.config_headers(data)
            if matrix_pins:
                revo.build_m_row_pins = matrix_pins[0]
                revo.build_m_col_pins = matrix_pins[1]  
                self.console.note(['Matrix pinout data found @ '+path])
                return
            else:
                continue

        self.console.warning(['Config.h header not found for '+self.build_kb.name, 'Matrix row/col pins must be provided manually!'], pause=True) 
        return

    def __get_keycodes(self, DEBUG=False):

        rev = self.build_rev.name
        revo = self.build_rev
        data = self.__cpp.preproc_keymap()
        token_list = _parse_txt.keymaps(data)
        function_token_list = _parse_txt.keymap_functions(data)

        layer_list = []
        layer_names = []
        num_col = 0
        layer_index = -1
        for tokens, start, stop in token_list:
            layer_index += 1
            if tokens.layer_name == '':
                name = str(layer_index)
                curr_layer = keycode_layer(name)
                layer_names.append(name)
            else:
                name = tokens.layer_name[1:-1]
                curr_layer = keycode_layer(tokens.layer_name[1:-1])
                layer_names.append(name)

            for row in tokens.layer:
                # Annoying fix for problem with QMK functions X(x,y)
                for i, element in enumerate(row):
                    if ')' in element and '(' not in element:
                        func = ''
                        func = row[i-1]+','+row[i] 
                        del row[i-1]
                        i -= 1
                        del row[i]
                        row.insert(i, func)
                if len(row) > curr_layer.matrix_cols:
                    num_col = len(row)
                curr_layer.keymap += (list(row))

            curr_layer.matrix_cols = num_col
            layer_list.append(curr_layer)

        # Functions
        functions = []
        for tokens, start, end in function_token_list:
            for f_token in tokens.function:

                if len(f_token) < 2:
                    index = None
                    function = f_token[0]
                else:
                    index = int(f_token[0])
                    function = f_token[1]

                if function in ref.keyp_actions.keys():
                    if index:
                        while len(functions) < index:
                            functions.append('')
                    functions.append(ref.keyp_actions[function])
                else:
                    if index:
                        while len(functions) < index:
                            functions.append('')
                    brac_index = function.index('(')+1
                    func = function[:brac_index]
                    target = function[brac_index:-1]

                    if func in ref.keyp_actions.keys() and target in layer_names:
                        layer = str(layer_names.index(target))
                        functions.append(ref.keyp_actions[func]+layer)
                    else:
                        functions.append('')

        if self.__args.debug or DEBUG:
            self.console.note(['Keycodes'])
            for layer in layer_list:
                print('Layer '+layer.name) 
                print(layer.keymap)

        if not layer_list:
            self.console.error(['Parsed and found no keymap', 'Failed to parse keymap file'])
        else:
            revo.build_layout = layer_list

        self.__convert_keycodes(layer_names, functions)

    def __convert_keycodes(self, layer_names, functions, DEBUG=False):

        for layer in self.build_rev.build_layout:
            if self.format == self.__output.keyplus:
                layer.convert_keyplus_keymap(layer_names, functions, self.console)
            #elif self.format == self.__output.kbfirmware:
                #layer.convert_kbf_keymap(layer_list, self.console)

    def __get_templates(self, DEBUG=False):

        rev = self.build_rev.name
        revo = self.build_rev

        if revo.template_list:
            keyboard_h = revo.template_loc

            with open(keyboard_h, 'r') as f:
                data = str(f.read())
            
            token_list = _parse_txt.layout_headers(data)
            for tokens, start, end in token_list:
                curr_template = layout_template(tokens.name)

                for row in tokens.layout:
                    layout_row = list(row)
                    curr_template.layout.append(layout_row)
                   
                array = list(tokens.array)
                for i, element in enumerate(array):
                    array[i] = re.sub('^[^##]*##', '', element) 

                curr_template.array = array
                revo.build_templates.append(curr_template)
            self.__convert_template_index()

        else:
            self.__generate_matrix_template()

        if self.__args.debug or DEBUG:
            self.console.note(['Templates'])
            for template in revo.build_templates:
                print(template.name)
                for row in template.layout:
                    print(row)
                print('Array')
                print(template.array)

    def __generate_matrix_template(self, index=0):

        rev = self.build_rev.name
        revo = self.build_rev
        layer = revo.build_layout[index]

        col_limit = layer.matrix_cols
        matrix = []
        for i in range(len(layer.keymap)):
            if i % col_limit == 0:
                if i >= col_limit:
                   matrix.append(row)
                row = []
                row.append(i)
            else:
                row.append(i)
        matrix.append(row)
        matrix_template = layout_template('!MATRIX LAYOUT')
        matrix_template.layout = matrix

        revo.build_templates.append(matrix_template)
        self.build_kb.build_template = '!MATRIX LAYOUT'

    def __convert_template_index(self):

        for template in self.build_rev.build_templates:
            template.convert_template_index(self.console, self.build_rev.build_templates)

    def __merge_layout_template(self, DEBUG=False):

        selected = self.build_kb.build_template
        layers = self.build_rev.build_layout
        templates = self.build_rev.build_templates

        for t in templates:
            if t.name == selected:
                template = t
                break

        self.console.note(['Building with template: '+selected])
        for x, layer in enumerate(layers):   

            layout_template = template.layout
            layout = copy.deepcopy(layout_template)
            keycode_array = layer.keymap
            max_index = len(keycode_array)

            try:
                for i, rows in enumerate(layout_template):
                    for j, ind in enumerate(rows):
                        if ind < max_index:
                            layout[i][j] = keycode_array[ind]
                        elif template.name != '!MATRIX LAYOUT':
                            self.console.warning(['Corrupt/incompatible layout template or keymap', 
                                'Invalid array value: '+str(ind),
                                'Trying again with default matrix layout...'])
                            self.__generate_matrix_template(x)
                            self.__merge_layout_template()
                            return
                        else:
                            self.console.error(['Corrupt/incompatible keymap', 'Invalid array value: '+str(ind)])
            except TypeError:
                self.console.error(['Corrupt layout template, invalid array index: '+str(ind)])

            if not self.build_rev.build_m_col_pins:
                col_limit = layer.matrix_cols
            else:
                col_limit = len(self.build_rev.build_m_col_pins)

            layer.layout = layout
            layer.matrix_map = layout_template
            if not self.__args.debug or DEBUG:
                self.console.note(['Layer '+layer.name])
            if DEBUG:
                layout_count = 0
                for row in layout:
                    layout_count += len(row)
                    print(str(len(row)) +'\t| '+str(row))
                print(layout_count)
                print('Matrix Map')
                array_count = 0
                for row in layout_template:
                    array_count += len(row)
                    print(str(len(row)) +'\t| '+str(row))
                print(array_count)
                print(str(max_index)+'\t| '+str(keycode_array))

    def __convert_matrix_map(self, DEBUG=False):
        layers = self.build_rev.build_layout

        for layer in layers:
            if not self.build_rev.build_m_col_pins:
                col_limit = layer.matrix_cols
            else:
                col_limit = len(self.build_rev.build_m_col_pins)
            if self.format == self.__output.keyplus:
                layer.convert_keyplus_matrix(col_limit)
            #elif self.format == self.__output.kbfirmware:
                #layer.convert_kbfirmware_matrix(col_limit)

            if self.__args.debug or DEBUG:
                self.console.note(['Layer '+layer.name])
                for row in layer.layout:
                    print(str(len(row)) +'\t| '+str(row))
                print('Array')
                for row in layer.matrix_map:
                    print(str(len(row)) +'\t| '+str(row))

    def __create_output(self, DEBUG=False):
        if self.format == self.__output.keyplus:
            self.__create_keyplus_yaml()
        #elif self.format == self.__output.kbfirmware:
            #layer.convert_kbfirmware_matrix(col_limit)

    def __create_keyplus_yaml(self, DEBUG=False):

        # Can't simply dump to yaml as we want to keep layout (array) as a human readable matrix (2D 'array').
        out_dir = self.dirs['Keyplus YAML output']
        kb_n = self.build_kb.name
        rev_n = self.build_rev.name
        if rev_n == 'n/a': rev_n = ''
        keymap = self.build_kb.build_keymap

        rev = self.build_rev
        layers = rev.build_layout
        errors = ''
        for error in self.console.errors:
            errors += '# '+error+'\n'

        template_matrix = layers[0].matrix_map
        if rev_n:
            name = kb_n+'_'+rev_n
        else:
            name = kb_n

        if rev.build_m_row_pins and rev.build_m_col_pins:
            rows = str(rev.build_m_row_pins)
            cols = str(rev.build_m_col_pins)
        else:
            rows = '# ------- Input row pins here ------- '
            cols = '# ------- Input col pins here ------- '
       
        template = ''
        for i, row in enumerate(template_matrix):
            for col in row:
                template += (col+', ')
            if i+1 < len(template_matrix):
                template += '\n        '

        layout = ''
        keycode_define = []
        for i, layer in enumerate(layers):
            layout += '      [ # layer '+str(i)+'\n        ['
            for row in layer.layout:
                layout += '\n          '
                for keycode in row:
                    if len(keycode) < 4:
                        repeat = 4 - len(keycode)
                        for i in (range(repeat)):
                            keycode+=' '
                    # tap>hold must be explicitly declared for now
                    if '>' in keycode and keycode != "'>' ":
                        if keycode not in keycode_define:
                            keycode_define.append(keycode)
                        #keycode = "'"+keycode+"'"

                    layout += keycode+', '
            layout +='\n        ]\n      ],\n'

        keycodes = ''
        if keycode_define:
            keycodes = 'keycodes:'
        for kc in keycode_define:
            split = kc.split('>', 1)
            tap = split[0][1:]
            hold = split[1][:-1]

            keycodes += ref.keyplus_yaml_keycode_template
            keycodes = keycodes.replace('<KEYCODE>', kc)
            keycodes = keycodes.replace('<TAP>', tap)
            keycodes = keycodes.replace('<HOLD>', hold)

        # Load Template
        output_yaml_info = ref.keyplus_yaml_template
        output_yaml_info = output_yaml_info.replace('<ERRORS>', errors)
        output_yaml_info = output_yaml_info.replace('<KB_NAME>', name)
        output_yaml_info = output_yaml_info.replace('<LAYOUT_NAME>', keymap)                              
        output_yaml_info = output_yaml_info.replace('<ROWS>', rows)                                 
        output_yaml_info = output_yaml_info.replace('<COLS>', cols)
        output_yaml_info = output_yaml_info.replace('<MATRIX_MAP>', template)
        output_yaml_info = output_yaml_info.replace('<LAYOUT>', layout)
        output_yaml_info = output_yaml_info.replace('<KEYCODES>', keycodes)

        kblibs = self.build_kb.libs
        if rev_n:
            path_list = kblibs + [rev_n, keymap]
        else:
            path_list = kblibs + [keymap]
        output_path = '_'.join(path_list)
        output_yaml = out_dir+output_path+'.yaml'
        if not os.path.exists(out_dir):
            try:
                os.makedirs(out_dir)
            except OSError as e:
                if e.errno != errno.EEXIST and os.path.isdir(out_dir):
                    raise
        try:
            with open(output_yaml, 'w') as f:
                f.write(output_yaml_info)
        except:
            self.console.error(['Failed to pipe output to '+output_yaml])

        if self.__args.debug or DEBUG:
            print(output_yaml_info)

        self.console.note(['SUCCESS! Output is in: '+output_yaml])


def q2keyplus():
    q2k = application('keyplus')
    q2k.execute()
'''
def q2kbfirmware():
    q2k = application('kbfirmware')
    q2k.execute()
'''

# Just here as an example for reference
'''
def q2keyplus_gui(keyboard, rev, keymap, template):
    q2k = application('keyplus', is_gui=True)
    q2k.set_kb(keyboard, rev, keymap, template)
    q2k.execute()
'''
#if __name__ == '__main__':
    #q2keyplus()
